import numpy as np
import casadi as cas
import rockit
from invariants_py import ocp_helper, dynamics_invariants, initialization
from invariants_py.ocp_helper import check_solver

class OCP_calc_pos:

    def __init__(self, window_len = 100, rms_error_traj = 10**-2, fatrop_solver = False, bool_unsigned_invariants = False):
        
        fatrop_solver = check_solver(fatrop_solver)               
        #%% Decision variables and parameters for the optimization problem 
        ocp = rockit.Ocp(T=1.0)
        N = window_len

        #% Decision variables

        # Define states
        p_obj = ocp.state(3,1) # object position
        R_t = ocp.state(3,3) # Frenet-Serret frame

        # Define controls
        invars = ocp.control(3) # three invariants [velocity | curvature rate | torsion rate]

        # Define parameters
        p_obj_m = ocp.parameter(3,grid='control+') # measured object positions
        #R_t_0 = ocp.parameter(3,3) # initial translational Frenet-Serret frame (for enforcing continuity of the moving frame)
        h = ocp.parameter(1) # stepsize
        
        # System dynamics (integrate current states + controls to obtain next states)
        (R_t_plus1, p_obj_plus1) = dynamics_invariants.vector_invariants_position(R_t, p_obj, invars, h)
        ocp.set_next(p_obj,p_obj_plus1)
        ocp.set_next(R_t,R_t_plus1)

        #% Constraints

        # Orthonormality of rotation matrix (only needed for one timestep, property is propagated by integrator)    
        ocp.subject_to(ocp.at_t0(ocp_helper.tril_vec(R_t.T @ R_t - np.eye(3))==0.))

        # Lower bounds on controls
        if bool_unsigned_invariants:
            ocp.subject_to(invars[0,:]>=0) # lower bounds on control
            ocp.subject_to(invars[1,:]>=0) # lower bounds on control

        # Measurement fitting constraint
        ek = cas.dot(p_obj - p_obj_m,p_obj - p_obj_m) # squared error

        running_ek = ocp.state() # running sum of squared error
        ocp.subject_to(ocp.at_t0(running_ek == 0))
        ocp.set_next(running_ek, running_ek + ek)

        total_ek = ocp.state() # total sum of squared error
        ocp.set_next(total_ek, total_ek)
        ocp.subject_to(ocp.at_tf(total_ek == running_ek + ek))
              
        total_ek_scaled = total_ek/N/rms_error_traj**2 # scaled total error
        ocp.subject_to(total_ek_scaled < 1)

        ocp.subject_to(ocp.at_t0(p_obj == p_obj_m)) # initial condition
        ocp.subject_to(ocp.at_tf(p_obj == p_obj_m)) # initial condition

        #% Objective function

        # Minimize moving frame invariants to deal with singularities and noise
        objective_reg = ocp.sum(cas.dot(invars[1:3],invars[1:3]))
        objective = objective_reg/(N-1)
        ocp.add_objective(objective)

        #%% Solver definition
        if fatrop_solver:
            ocp.method(rockit.external_method('fatrop', N=N-1))
            ocp._method.set_name("/codegen/position") # pick a unique name when using multiple OCP specifications in the same script 
            # import random
            # import string
            # rand = "".join(random.choices(string.ascii_lowercase))
            # ocp._method.set_name("/codegen/calculate_position_"+rand)          
        else:
            ocp.method(rockit.MultipleShooting(N=N-1))
            ocp.solver('ipopt', {'expand':True, 'ipopt.print_info_string':'yes'})
            # ocp.solver('ipopt',{"print_time":True,"expand":True},{'gamma_theta':1e-12,'max_iter':200,'tol':1e-4,'print_level':5,'ma57_automatic_scaling':'no','linear_solver':'mumps'})
        
        # Solve already once with dummy values for code generation (TODO: can this step be avoided somehow?)
        ocp.set_initial(R_t, np.eye(3))
        ocp.set_initial(invars, np.array([1,0.01,0.01]))
        ocp.set_value(p_obj_m, np.vstack((np.linspace(0, 1, N), np.ones((2, N)))))
        ocp.set_value(h, 0.01)
        ocp.solve_limited() # code generation

        # Set solver options (TODO: why can this not be done before solving?)
        if fatrop_solver:
            ocp._method.set_option("tol",1e-6)
            #ocp._method.set_option("print_level",0)
        self.first_time = True
        
        # Encapsulate whole rockit specification in a casadi function
        p_obj_m_sampled = ocp.sample(p_obj_m, grid='control')[1] # sampled measured object positions
        h_value = ocp.value(h) # value of stepsize
        solution = [ocp.sample(invars, grid='control-')[1],
            ocp.sample(p_obj, grid='control')[1], # sampled object positions
            ocp.sample(R_t, grid='control')[1]] # sampled FS frame

        self.ocp = ocp # save the optimization problem locally, avoids problems when multiple rockit ocp's are created
        self.ocp_function = self.ocp.to_function('ocp_function', 
            [p_obj_m_sampled,h_value,*solution], # inputs
            [*solution], # outputs
            ["p_obj_m","h","invars1","p_obj1","R_t_1"], # input labels for debugging
            ["invars2","p_obj2","R_t_2"], # output labels for debugging
        )

    def calculate_invariants(self,measured_positions,stepsize):
        # Check if this is the first function call
        if self.first_time:
            # Initialize states and controls using measurements
            self.solution,measured_positions = initialization.initialize_VI_pos(measured_positions)
            self.first_time = False

        # Solve the optimization problem for the given measurements starting from previous solution
        self.solution = self.ocp_function(measured_positions.T, stepsize, *self.solution)

        # Return the results    
        invars, p_obj_sol, R_t_sol  = self.solution # unpack the results            
        invariants = np.array(invars).T # make a N-1 x 3 array
        invariants = np.vstack((invariants, invariants[-1,:])) # make a N x 3 array by repeating last row
        calculated_trajectory = np.array(p_obj_sol).T # make a N x 3 array
        calculated_movingframe = np.transpose(np.reshape(R_t_sol.T, (-1, 3, 3)), (0, 2, 1))
        return invariants, calculated_trajectory, calculated_movingframe
    
if __name__ == "__main__":
    # Test data
    measured_positions = np.array([[1, 2, 3], [4.1, 5.5, 6], [7, 8.5, 9], [10, 11.9, 12]])
    timestep = 0.05
    
    # Specify OCP symbolically
    N = np.size(measured_positions,0)
    OCP = OCP_calc_pos(window_len=N,fatrop_solver=False)

    # Solve the OCP using the specified data
    calc_invariants, calc_trajectory, calc_movingframes = OCP.calculate_invariants(measured_positions, timestep)

    # Reuse the OCP using new data (for example in moving horizon estimation)
    # TODO if measured_positions is used then it does not converge in one iteration because lam/mu is not taken from previous solution
    measured_positions_2 = np.array([[1.2, 2.3, 3.4], [4.5, 5.6, 6.7], [7.8, 8.9, 9.0], [10.1, 11.2, 12.3]])
    calc_invariants, calc_trajectory, calc_movingframes = OCP.calculate_invariants(measured_positions_2, timestep)

    # Print the results
    #print("Calculated invariants:")
    #print(calc_invariants)
    #print(np.size(calc_invariants))
    #print("Calculated Moving Frame:")
    print(calc_movingframes)
    #print("Calculated Trajectory:")
    print(calc_trajectory)


"""
Simple test to calculate invariants using opti specification
"""

import numpy as np
import invariants_py.reparameterization as reparam
import invariants_py.opti_calculate_vector_invariants_position_mj as FS3
import invariants_py.plotters as plotters
import os
import time
from invariants_py import data_handler as dh

#%% Settings

weight_measurements = 100
weight_regularization = 10**-10


#%% Load data

# load data
data_location = dh.find_data_path("contour_coordinates.out")
position_data = np.loadtxt(data_location, dtype='float')

# reparameterize to arc length
trajectory,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_positiontrajectory_arclength(position_data)

# plot data
plotters.plot_2D_contour(trajectory)


#%% Calculate invariants

# specify optimization problem symbolically
FS_calculation_problem = FS3.OCP_calc_pos(window_len=nb_samples, w_pos = weight_measurements, w_regul = weight_regularization)

# calculate invariants given measurements
start_time = time.time()
invariants, trajectory_recon, movingframes = FS_calculation_problem.calculate_invariants_global(trajectory,stepsize)
end_time = time.time()
print('')
print("Total time solving + sampling overhead: ")
print(end_time - start_time)

# figures
plotters.plot_trajectory_invariants(trajectory,trajectory_recon,arclength,invariants)




import numpy as np
from math import pi
import casadi as cas
import invariants_py.dynamics_invariants as dynamics

class OCP_calc_rot:

    def __init__(self, window_len = 100, bool_unsigned_invariants = False, rms_error_traj = 2*pi/180):
       
        #%% Create decision variables and parameters for the optimization problem
        
        opti = cas.Opti() # use OptiStack package from Casadi for easy bookkeeping of variables (no cumbersome indexing)

        # Define system states X (unknown object pose + moving frame pose at every time step) 
        R_obj = []
        R_r = []
        X = []
        for k in range(window_len):
            R_obj.append(opti.variable(3,3)) # object orientation
            R_r.append(opti.variable(3,3)) # rotational Frenet-Serret frame
            X.append(cas.vertcat(cas.vec(R_r[k]), cas.vec(R_obj[k])))

        # Define system controls (invariants at every time step)
        U = opti.variable(3,window_len-1)

        # Define system parameters P (known values in optimization that need to be set right before solving)
        R_obj_m = [] # measured object orientation
        R_r_0 = opti.parameter(3,3) # THIS IS COMMENTED OUT IN MATLAB, WHY?
        for k in range(window_len):
            R_obj_m.append(opti.parameter(3,3)) # object orientation

        h = opti.parameter(1,1)
        
        #%% Specifying the constraints
        
        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        #opti.subject_to( R_t[-1].T @ R_t[-1] == np.eye(3))
        
        constr = R_r[0].T @ R_r[0]
        opti.subject_to(cas.vertcat(constr[1,0], constr[2,0], constr[2,1], constr[0,0], constr[1,1], constr[2,2])==cas.vertcat(cas.DM.zeros(3,1), cas.DM.ones(3,1)))

        constr_obj = R_obj[0].T @ R_obj[0]
        opti.subject_to(cas.vertcat(constr_obj[1,0], constr_obj[2,0], constr_obj[2,1], constr_obj[0,0], constr_obj[1,1], constr_obj[2,2])==cas.vertcat(cas.DM.zeros(3,1), cas.DM.ones(3,1)))


        # Dynamic constraints
        integrator = dynamics.define_geom_integrator_rot_FSI_casadi(h)
        for k in range(window_len-1):
            # Integrate current state to obtain next state (next rotation and position)
            Xk_end = integrator(X[k],U[:,k],h)
            
            # Gap closing constraint
            opti.subject_to(Xk_end==X[k+1])
            
        # Lower bounds on controls
        if bool_unsigned_invariants:
            opti.subject_to(U[0,:]>=0) # lower bounds on control
            opti.subject_to(U[1,:]>=0) # lower bounds on control

        #%% Specifying the objective

        # Fitting constraint to remain close to measurements
        objective_fit = 0
        for k in range(window_len):
            err_rot = R_obj_m[k].T @ R_obj[k] - np.eye(3) # orientation error
            objective_fit = objective_fit + cas.dot(err_rot,err_rot)
            
        opti.subject_to(objective_fit/window_len/rms_error_traj**2 < 1)

        # Regularization constraints to deal with singularities and noise
        objective_reg = 0
        for k in range(window_len-1):
            #if k!=0:
            #    err_deriv = U[:,k] - U[:,k-1] # first-order finite backwards derivative (noise smoothing effect)
            #else:
            #    err_deriv = 0
            err_abs = U[[1,2],k] # absolute value invariants (force arbitrary invariants to zero)

            objective_reg = objective_reg + cas.dot(err_abs,err_abs)

            ##Check that obj function is correctly typed in !!!
            #objective_reg = objective_reg \
            #                + cas.dot(w_deriv**(0.5)*err_deriv,w_deriv**(0.5)*err_deriv) \
            #                + cas.dot(w_abs**(0.5)*err_abs, w_abs**(0.5)*err_abs)

        objective = objective_reg/(window_len-1)

        opti.subject_to(U[1,-1] == U[1,-2]); # Last sample has no impact on RMS error

        #%% Define solver and save variables
        opti.minimize(objective)
        opti.solver('ipopt',{"print_time":True,"expand":True},{'gamma_theta':1e-12,'max_iter':200,'tol':1e-4,'print_level':5,'ma57_automatic_scaling':'no','linear_solver':'mumps'})
        
        # Save variables
        self.R_r = R_r
        self.R_obj = R_obj
        self.U = U
        self.R_obj_m = R_obj_m
        self.R_r_0 = R_r_0
        self.window_len = window_len
        self.opti = opti
        self.first_window = True
        self.h = h
         
    def calculate_invariants_global(self,trajectory_meas,stepsize): 
        #%%

        measured_orientation = trajectory_meas[:,:3,:3]
        N = self.window_len
        
        # Initialize states
        # TODO  this is not correct yet, ex not perpendicular to ey
        # Pdiff = np.diff(measured_orientation,axis=0)
        # ex = Pdiff / np.linalg.norm(Pdiff,axis=1).reshape(N-1,1)
        # ex = np.vstack((ex,[ex[-1,:]]))
        # ey = np.tile( np.array((0,0,1)), (N,1) )
        # ez = np.array([np.cross(ex[i,:],ey[i,:]) for i in range(N)])
        
        #JUST TESTING
        ex = np.tile( np.array((1,0,0)), (N,1) )
        ey = np.tile( np.array((0,1,0)), (N,1) )
        ez = np.array([np.cross(ex[i,:],ey[i,:]) for i in range(N)])
        
        #Pdiff_cross = np.cross(Pdiff[0:-1],Pdiff[1:])
        #ey = Pdiff_cross / np.linalg.norm(Pdiff_cross,axis=1).reshape(N-2,1)
        
        for k in range(N):
            self.opti.set_initial(self.R_r[k], np.array([ex[k,:], ey[k,:], ez[k,:]]).T ) #construct_init_FS_from_traj(meas_traj.Obj_location)
            self.opti.set_initial(self.R_obj[k], measured_orientation[k])
            
        # Initialize controls
        for k in range(N-1):    
            self.opti.set_initial(self.U[:,k], np.array([1,1e-12,1e-12]))
            
        # Set values parameters
        self.opti.set_value(self.R_r_0, np.eye(3,3)) # THIS IS COMMENTED OUT IN MATLAB, WHY?
        for k in range(N):
            self.opti.set_value(self.R_obj_m[k], measured_orientation[k])       
        
        self.opti.set_value(self.h,stepsize)

        # ######################
        # ##  DEBUGGING: check integrator in initial values, time step 0 to 1
        # x0 = cas.vertcat(cas.vec(np.eye(3,3)), cas.vec(measured_positions[0]))
        # u0 = 1e-8*np.ones((3,1))
        # integrator = dynamics.define_geom_integrator_tra_FSI_casadi(self.stepsize)
        # x1 = integrator(x0,u0)
        # print(x1)
        # ######################

        # Solve the NLP
        sol = self.opti.solve_limited()
        self.sol = sol
        
        # Extract the solved variables
        invariants = sol.value(self.U).T
        invariants =  np.vstack((invariants,[invariants[-1,:]]))
        calculated_trajectory = np.array([sol.value(i) for i in self.R_obj])
        calculated_movingframe = np.array([sol.value(i) for i in self.R_r])
        
        return invariants, calculated_trajectory, calculated_movingframe

    def calculate_invariants_online(self,trajectory_meas,stepsize,sample_jump):
        #%%
        if self.first_window:
            # Calculate invariants in first window
            invariants, calculated_trajectory, calculated_movingframe = self.calculate_invariants_global(trajectory_meas,stepsize)
            self.first_window = False
            
            
            # Add continuity constraints on first sample
            #self.opti.subject_to( self.R_t[0] == self.R_t_0 )
            #self.opti.subject_to( self.p_obj[0] == self.p_obj_m[0])
            
            return invariants, calculated_trajectory, calculated_movingframe
        else:
            
            measured_orientation = trajectory_meas[:,:3,:3]
            N = self.window_len
            
            #%% Set values parameters
            #for k in range(1,N):
            #    self.opti.set_value(self.p_obj_m[k], measured_positions[k-1])   
            
            for k in range(0,N):
                    self.opti.set_value(self.R_obj_m[k], measured_orientation[k])   
            
            # Set other parameters equal to the measurements in that window
            self.opti.set_value(self.R_r_0, self.sol.value(self.R_r[sample_jump]))
            #self.opti.set_value(self.R_obj_m[0], self.sol.value(self.R_obj[sample_jump]))
            
            self.opti.set_value(self.h,stepsize)
        
            #%% First part of window initialized using results from earlier solution
            # Initialize states
            for k in range(N-sample_jump-1):
                self.opti.set_initial(self.R_r[k], self.sol.value(self.R_r[sample_jump+k]))
                self.opti.set_initial(self.R_obj[k], self.sol.value(self.R_obj[sample_jump+k]))
                
            # Initialize controls
            for k in range(N-sample_jump-1):    
                self.opti.set_initial(self.U[:,k], self.sol.value(self.U[:,sample_jump+k]))
                
            #%% Second part of window initialized uses default initialization
            # Initialize states
            for k in range(N-sample_jump,N):
                self.opti.set_initial(self.R_r[k], self.sol.value(self.R_r[-1]))
                self.opti.set_initial(self.R_obj[k], measured_orientation[k-1])
                
            # Initialize controls
            for k in range(N-sample_jump-1,N-1):    
                self.opti.set_initial(self.U[:,k], 1e-3*np.ones((3,1)))

            #print(self.sol.value(self.R_t[-1]))

            #%% Solve the NLP
            sol = self.opti.solve_limited()
            self.sol = sol
            
            # Extract the solved variables
            invariants = sol.value(self.U).T
            invariants =  np.vstack((invariants,[invariants[-1,:]]))
            calculated_trajectory = np.array([sol.value(i) for i in self.p_obj])
            calculated_movingframe = np.array([sol.value(i) for i in self.R_t])
            
            return invariants, calculated_trajectory, calculated_movingframe


"""
Application involving approach trajectory of a bottle opener towards a bottle
"""

# Imports
import numpy as np
from math import pi
import os
import invariants_py.data_handler as dh
import matplotlib.pyplot as plt
import invariants_py.reparameterization as reparam
import scipy.interpolate as ip

from invariants_py.rockit_calculate_vector_invariants_position import OCP_calc_pos
from invariants_py.rockit_calculate_vector_invariants_rotation import OCP_calc_rot
from invariants_py.rockit_generate_pose_from_vector_invariants import OCP_gen_pose
from invariants_py.rockit_generate_pose_from_vector_invariants_jointlim import OCP_gen_pose_jointlim
from scipy.spatial.transform import Rotation as R
from invariants_py.orthonormalize_rotation import orthonormalize_rotation as orthonormalize
import invariants_py.plotters as pl
import collision_detection_bottle as cd
from invariants_py.reparameterization import interpR
from invariants_py.initialization import FSr_init

#%%
data_location = dh.find_data_path('beer_1.txt')
trajectory,time = dh.read_pose_trajectory_from_txt(data_location)
pose,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_trajectory_arclength(trajectory)
arclength_n = arclength/arclength[-1]
home_pos = [0,0,0] # Use this if not considering the robot
home_pos = [0.3056, 0.0635, 0.441] # Define home position of the robot
trajectory_position = pose[:,:3,3] + home_pos
trajectory_orientation = pose[:,:3,:3]

fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111, projection='3d')
ax = plt.axes(projection='3d')
ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'.-')
n_frames = 10
indx = np.trunc(np.linspace(0,len(trajectory_orientation)-1,n_frames))
indx = indx.astype(int)
opener_location = dh.find_data_path('opener.stl')
bottle_location = dh.find_data_path('bottle.stl')
for i in indx:
    pl.plot_3d_frame(trajectory_position[i,:],trajectory_orientation[i,:,:],1,0.01,['red','green','blue'],ax)
    pl.plot_stl(opener_location,trajectory_position[i,:],trajectory_orientation[i,:,:],colour="c",alpha=0.2,ax=ax)    

#%%
# define class for OCP results
class OCP_results:

    def __init__(self,FSt_frames,FSr_frames,Obj_pos,Obj_frames,invariants):
        self.FSt_frames = FSt_frames
        self.FSr_frames = FSr_frames
        self.Obj_pos = Obj_pos
        self.Obj_frames = Obj_frames
        self.invariants = invariants

optim_calc_results = OCP_results(FSt_frames = [], FSr_frames = [], Obj_pos = [], Obj_frames = [], invariants = np.zeros((len(trajectory),6)))

# choose solver
use_fatrop_solver = True # True = fatrop, False = ipopt

# specify optimization problem symbolically
FS_calculation_problem_pos = OCP_calc_pos(window_len=nb_samples, bool_unsigned_invariants = False, rms_error_traj = 0.004, fatrop_solver = use_fatrop_solver)
FS_calculation_problem_rot = OCP_calc_rot(window_len=nb_samples, bool_unsigned_invariants = False, rms_error_traj = 2*pi/180, fatrop_solver = use_fatrop_solver) 

# calculate invariants given measurements
optim_calc_results.invariants[:,3:], optim_calc_results.Obj_pos, optim_calc_results.FSt_frames = FS_calculation_problem_pos.calculate_invariants(trajectory,stepsize)

optim_calc_results.invariants[:,:3], optim_calc_results.Obj_frames, optim_calc_results.FSr_frames = FS_calculation_problem_rot.calculate_invariants(trajectory,stepsize)
optim_calc_results.Obj_pos += home_pos
fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111, projection='3d')
ax = plt.axes(projection='3d')
ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'b')
ax.plot(optim_calc_results.Obj_pos[:,0],optim_calc_results.Obj_pos[:,1],optim_calc_results.Obj_pos[:,2],'r')
indx = np.trunc(np.linspace(0,len(optim_calc_results.Obj_pos)-1,n_frames))
indx = indx.astype(int)
for i in indx:
    pl.plot_3d_frame(optim_calc_results.Obj_pos[i,:],optim_calc_results.Obj_frames[i,:,:],1,0.01,['red','green','blue'],ax)
    pl.plot_stl(opener_location,trajectory_position[i,:],trajectory_orientation[i,:,:],colour="c",alpha=0.2,ax=ax)
    pl.plot_stl(opener_location,optim_calc_results.Obj_pos[i,:],optim_calc_results.Obj_frames[i,:,:],colour="r",alpha=0.2,ax=ax)

pl.plot_orientation(optim_calc_results.Obj_frames,trajectory_orientation)

pl.plot_invariants(optim_calc_results.invariants,[],arclength_n)

if plt.get_backend() != 'agg':
    plt.show()

#%%
# Spline of model
knots = np.concatenate(([arclength_n[0]],[arclength_n[0]],arclength_n,[arclength_n[-1]],[arclength_n[-1]]))
degree = 3
spline_model_trajectory = ip.BSpline(knots,optim_calc_results.invariants,degree)

def interpolate_model_invariants(demo_invariants, progress_values):
    
    resampled_invariants = np.array([demo_invariants(i) for i in progress_values]) 
    new_stepsize = progress_values[1] - progress_values[0] 
    
    resampled_invariants[:,0] = resampled_invariants[:,0] *  (progress_values[-1] - progress_values[0])
    return resampled_invariants, new_stepsize

#%% 
current_progress = 0
number_samples = 100

progress_values = np.linspace(current_progress, arclength_n[-1], number_samples)
model_invariants,new_stepsize = interpolate_model_invariants(spline_model_trajectory,progress_values)

pl.plot_interpolated_invariants(optim_calc_results.invariants, model_invariants, arclength_n, progress_values)

# new constraints
current_index = round(current_progress*len(trajectory))
p_obj_start = optim_calc_results.Obj_pos[current_index]
R_obj_start = orthonormalize(optim_calc_results.Obj_frames[current_index])
FSt_start = orthonormalize(optim_calc_results.FSt_frames[current_index])
# FSr_start = orthonormalize(optim_calc_results.FSr_frames[current_index])
p_obj_end = optim_calc_results.Obj_pos[-1] + np.array([0.1,0.1,0.1])
alpha = 30
rotate = R.from_euler('z', alpha, degrees=True)
R_obj_end =  orthonormalize(rotate.as_matrix() @ optim_calc_results.Obj_frames[-1])
FSt_end = orthonormalize(rotate.as_matrix() @ optim_calc_results.FSt_frames[-1])
# FSr_end = orthonormalize(optim_calc_results.FSr_frames[-1])

# define new class for OCP results
optim_gen_results = OCP_results(FSt_frames = [], FSr_frames = [], Obj_pos = [], Obj_frames = [], invariants = np.zeros((number_samples,6)))

# Robot urdf location
path_to_urdf = dh.find_data_path('ur10.urdf')

# specify optimization problem symbolically
# FS_online_generation_problem = OCP_gen_pose(window_len=number_samples, fatrop_solver = use_fatrop_solver)
FS_online_generation_problem = OCP_gen_pose_jointlim(path_to_urdf = path_to_urdf, tip = 'TCP_frame', window_len=number_samples, fatrop_solver = use_fatrop_solver)

# Linear initialization
R_obj_init = interpR(np.linspace(0, 1, len(optim_calc_results.Obj_frames)), [0,1], np.array([R_obj_start, R_obj_end]))
# R_r_init = interpR(np.linspace(0, 1, len(optim_calc_results.FSr_frames)), [0,1], np.array([FSr_start, FSr_end]))

R_r_init, R_r_init_array, U_init = FSr_init(R_obj_start, R_obj_end)

# Define OCP weights
w_invars = np.array([1, 1, 1, 5*10**1, 1.0, 1.0]) # i_r1, i_r2, i_r3, i_t1, i_t2, i_t3
w_high_active = 0
w_high_start = 60
w_high_end = number_samples
w_invars_high = 10*w_invars

# Join initialization and limits
q_init = [-pi, -2.27, 2.27, -pi/2, -pi/2, pi/4] * np.ones((number_samples,6))
q_joint_lim = [2*pi, 2*pi, pi, 2*pi, 2*pi, 2*pi]
# Solve
# optim_gen_results.invariants, optim_gen_results.Obj_pos, optim_gen_results.Obj_frames, optim_gen_results.FSt_frames, optim_gen_results.FSr_frames, tot_time = FS_online_generation_problem.generate_trajectory(U_demo = model_invariants, p_obj_init = optim_calc_results.Obj_pos, R_obj_init = R_obj_init, R_t_init = optim_calc_results.FSt_frames, R_r_init = R_r_init_array, R_t_start = FSt_start, R_r_start = R_r_init, R_t_end = FSt_end, R_r_end = R_r_init, p_obj_start = p_obj_start, R_obj_start = R_obj_start, p_obj_end = p_obj_end, R_obj_end = R_obj_end, step_size = new_stepsize, w_high_start = w_high_start, w_high_end = w_high_end, w_high_invars = w_invars_high, w_invars = w_invars, w_high_active = w_high_active)
optim_gen_results.invariants, optim_gen_results.Obj_pos, optim_gen_results.Obj_frames, optim_gen_results.FSt_frames, optim_gen_results.FSr_frames, tot_time, joint_values = FS_online_generation_problem.generate_trajectory(U_demo = model_invariants, p_obj_init = optim_calc_results.Obj_pos, R_obj_init = R_obj_init, R_t_init = optim_calc_results.FSt_frames, R_r_init = R_r_init_array, q_init = q_init, q_lim = q_joint_lim, R_t_start = FSt_start, R_r_start = R_r_init, R_t_end = FSt_end, R_r_end = R_r_init, p_obj_start = p_obj_start, R_obj_start = R_obj_start, p_obj_end = p_obj_end, R_obj_end = R_obj_end, step_size = new_stepsize, w_high_start = w_high_start, w_high_end = w_high_end, w_high_invars = w_invars_high, w_invars = w_invars, w_high_active = w_high_active)

if use_fatrop_solver:
    print('')
    print("TOTAL time to generate new trajectory: ")
    print(str(tot_time) + "[s]")

# print('Joint values:')
# print(joint_values)

# optim_gen_results.Obj_frames = interpR(np.linspace(0, 1, len(optim_calc_results.Obj_frames)), [0,1], np.array([R_obj_start, R_obj_end])) # JUST TO CHECK INITIALIZATION

for i in range(len(optim_gen_results.Obj_frames)):
    optim_gen_results.Obj_frames[i] = orthonormalize(optim_gen_results.Obj_frames[i])

fig = plt.figure(figsize=(14,8))
ax = fig.add_subplot(111, projection='3d')
ax.plot(optim_calc_results.Obj_pos[:,0],optim_calc_results.Obj_pos[:,1],optim_calc_results.Obj_pos[:,2],'b')
ax.plot(optim_gen_results.Obj_pos[:,0],optim_gen_results.Obj_pos[:,1],optim_gen_results.Obj_pos[:,2],'r')
for i in indx:
    pl.plot_stl(opener_location,optim_calc_results.Obj_pos[i,:],optim_calc_results.Obj_frames[i,:,:],colour="c",alpha=0.2,ax=ax)

indx_online = np.trunc(np.linspace(0,len(optim_gen_results.Obj_pos)-1,n_frames))
indx_online = indx_online.astype(int)
for i in indx_online:
    pl.plot_3d_frame(optim_calc_results.Obj_pos[i,:],optim_calc_results.Obj_frames[i,:,:],1,0.01,['red','green','blue'],ax)
    pl.plot_3d_frame(optim_gen_results.Obj_pos[i,:],optim_gen_results.Obj_frames[i,:,:],1,0.01,['red','green','blue'],ax)
    pl.plot_stl(opener_location,optim_gen_results.Obj_pos[i,:],optim_gen_results.Obj_frames[i,:,:],colour="r",alpha=0.2,ax=ax)
r_bottle = 0.0145+0.01 # bottle radius + margin
obj_pos = p_obj_end + [r_bottle*np.sin(alpha*pi/180) , -r_bottle*np.cos(alpha*pi/180), 0] # position of the bottle
pl.plot_stl(bottle_location,obj_pos,np.eye(3),colour="tab:gray",alpha=1,ax=ax)
pl.plot_orientation(optim_calc_results.Obj_frames,optim_gen_results.Obj_frames,current_index)

pl.plot_invariants(optim_calc_results.invariants, optim_gen_results.invariants, arclength_n, progress_values)

fig99 = plt.figure(figsize=(14,8))
ax99 = fig99.add_subplot(111, projection='3d')
pl.plot_stl(opener_location,[0,0,0],optim_calc_results.Obj_frames[-1],colour="r",alpha=0.5,ax=ax99)
pl.plot_stl(opener_location,[0,0,0],R_obj_end,colour="b",alpha=0.5,ax=ax99)
pl.plot_stl(opener_location,[0,0,0],optim_gen_results.Obj_frames[-1],colour="g",alpha=0.5,ax=ax99)

opener_dim_x = 0.04
opener_dim_y = 0.15
opener_dim_z = 0
opener_points = 30
offset = -0.02 # position of the hook where have contact with bottle
opener_geom = np.zeros((opener_points,3))
for j in range(opener_points // 3):
    opener_geom[j*3, :] = [opener_dim_x/2, offset-j*offset, opener_dim_z]
    opener_geom[j*3+1, :] = [0, offset-j*offset, opener_dim_z]
    opener_geom[j*3+2, :] = [-opener_dim_x/2, offset-j*offset, opener_dim_z]

tilting_angle_rotx_deg=0
tilting_angle_roty_deg=0
tilting_angle_rotz_deg=0
mode = 'rpy'
collision_flag, first_collision_sample, last_collision_sample = cd.collision_detection_bottle(optim_gen_results.Obj_pos,optim_gen_results.Obj_frames,obj_pos,opener_geom,tilting_angle_rotx_deg,tilting_angle_roty_deg,tilting_angle_rotz_deg,mode,ax)

if collision_flag:
    print("COLLISION DETECTED")
    print("First collision sample: " + str(first_collision_sample))
    print("Last collision sample: " + str(last_collision_sample))
else:
    print("NO COLLISION DETECTED")

if plt.get_backend() != 'agg':
    plt.show()


# # Generation of multiple trajectories to test FATROP calculation speed

# current_progress = 0
# number_samples = 100
# number_of_trajectories = 100

# progress_values = np.linspace(current_progress, arclength_n[-1], number_samples)
# model_invariants,new_stepsize = interpolate_model_invariants(spline_model_trajectory,progress_values)

# # pl.plot_interpolated_invariants(optim_calc_results.invariants, model_invariants, arclength_n, progress_values)

# # new constraints
# current_index = round(current_progress*len(trajectory))
# p_obj_start = optim_calc_results.Obj_pos[current_index]
# R_obj_start = orthonormalize(optim_calc_results.Obj_frames[current_index])
# FSt_start = orthonormalize(optim_calc_results.FSt_frames[current_index])
# FSr_start = orthonormalize(optim_calc_results.FSr_frames[current_index])
# FSt_end = orthonormalize(optim_calc_results.FSt_frames[-1])
# FSr_end = orthonormalize(optim_calc_results.FSr_frames[-1])

# # define new class for OCP results
# optim_gen_results = OCP_results(FSt_frames = [], FSr_frames = [], Obj_pos = [], Obj_frames = [], invariants = np.zeros((number_samples,6)))

# # specify optimization problem symbolically
# FS_online_generation_problem = OCP_gen_pose_jointlim(path_to_urdf = path_to_urdf, tip = 'TCP_frame', window_len=number_samples, fatrop_solver = use_fatrop_solver)

# fig = plt.figure(figsize=(14,8))
# ax = fig.add_subplot(111, projection='3d')
# ax.plot(optim_calc_results.Obj_pos[:,0],optim_calc_results.Obj_pos[:,1],optim_calc_results.Obj_pos[:,2],'b')

# total_time = 0
# counter = 0
# max_time = 0
# targets = np.zeros((number_of_trajectories,4))
# for k in range(len(targets)):
# # for x in range(-2,3):
#     # for y in range(-2,3):
#         # p_obj_end = optim_calc_results.Obj_pos[-1] + np.array([0.05*x,0.05*y,0])
#     targets[k,:-1] = optim_calc_results.Obj_pos[-1] + np.array([random.uniform(-0.2,0.2),random.uniform(-0.2,0.2),random.uniform(-0.05,0.05)])
#     targets[k,-1] = random.uniform(0,30)
#     p_obj_end = targets[k,:-1]
#     rotate = R.from_euler('z', targets[k,-1], degrees=True)
#     R_obj_end =  orthonormalize(rotate.apply(optim_calc_results.Obj_frames[-1]))
    
#     # Solve
#     optim_gen_results.invariants, optim_gen_results.Obj_pos, optim_gen_results.Obj_frames, optim_gen_results.FSt_frames, optim_gen_results.FSr_frames, tot_time, joint_values = FS_online_generation_problem.generate_trajectory(U_demo = model_invariants, p_obj_init = optim_calc_results.Obj_pos, R_obj_init = R_obj_init, R_t_init = optim_calc_results.FSt_frames, R_r_init = R_r_init_array, q_init = q_init, q_lim = q_joint_lim, R_t_start = FSt_start, R_r_start = R_r_init, R_t_end = FSt_end, R_r_end = R_r_init, p_obj_start = p_obj_start, R_obj_start = R_obj_start, p_obj_end = p_obj_end, R_obj_end = R_obj_end, step_size = new_stepsize, w_high_start = w_high_start, w_high_end = w_high_end, w_high_invars = w_invars_high, w_invars = w_invars, w_high_active = w_high_active)
    
#     for i in range(len(optim_gen_results.Obj_frames)):
#         optim_gen_results.Obj_frames[i] = orthonormalize(optim_gen_results.Obj_frames[i])

#     ax.plot(optim_gen_results.Obj_pos[:,0],optim_gen_results.Obj_pos[:,1],optim_gen_results.Obj_pos[:,2],'r')

#     # indx_online = np.trunc(np.linspace(0,len(optim_gen_results.Obj_pos)-1,n_frames))
#     # indx_online = indx_online.astype(int)
#     # for i in indx_online:
#     #     pl.plot_3d_frame(optim_calc_results.Obj_pos[i,:],optim_calc_results.Obj_frames[i,:,:],1,0.01,['red','green','blue'],ax)
#     #     pl.plot_3d_frame(optim_gen_results.Obj_pos[i,:],optim_gen_results.Obj_frames[i,:,:],1,0.01,['red','green','blue'],ax)
#         # pl.plot_stl(opener_location,optim_gen_results.Obj_pos[i,:],optim_gen_results.Obj_frames[i,:,:],colour="r",alpha=0.2,ax=ax)
#     if use_fatrop_solver:
#         new_time = tot_time
#         if new_time > max_time:
#             max_time = new_time
#         total_time += new_time
    
#     counter += 1
#     # plt.show(block=False)
# if use_fatrop_solver:
#     print('')
#     print("AVERAGE time to generate new trajectory: ")
#     print(str(total_time/counter) + "[s]") # last time average was 80ms
#     print('')
#     print("MAXIMUM time to generate new trajectory: ")
#     print(str(max_time) + "[s]") # last time maximum was 156ms

# # fig = plt.figure(figsize=(10,6))
# # ax1 = fig.add_subplot(111, projection='3d')
# # ax1 = plt.axes(projection='3d')
# # ax1.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'b')
# # ax1.plot(targets[:,0],targets[:,1],targets[:,2],'r.')

# fig = plt.figure(figsize=(5,5))
# ax2 = fig.add_subplot()
# ax2.plot(targets[:,-1],'r.')

# if plt.get_backend() != 'agg':
#    plt.show()




# Imports
import numpy as np
from math import pi
import invariants_py.data_handler as dh
import matplotlib.pyplot as plt
import invariants_py.reparameterization as reparam
import scipy.interpolate as ip
from invariants_py.rockit_calculate_vector_invariants_position import OCP_calc_pos as FS_calc_pos
from invariants_py.rockit_calculate_vector_invariants_rotation import OCP_calc_rot as FS_calc_rot
from invariants_py.rockit_generate_position_from_vector_invariants import OCP_gen_pos as FS_gen_pos
from invariants_py.rockit_generate_rotation_from_vector_invariants import OCP_gen_rot as FS_gen_rot
from IPython.display import clear_output
from scipy.spatial.transform import Rotation as R
from invariants_py.orthonormalize_rotation import orthonormalize_rotation as orthonormalize
import invariants_py.plotters as pl
from invariants_py.reparameterization import interpR
from invariants_py.initialization import FSr_init

# define class for OCP results
class OCP_results:

    def __init__(self,FSt_frames,FSr_frames,Obj_pos,Obj_frames,invariants):
        self.FSt_frames = FSt_frames
        self.FSr_frames = FSr_frames
        self.Obj_pos = Obj_pos
        self.Obj_frames = Obj_frames
        self.invariants = invariants
#%%
def calculate_invariants(data_location, plot_demo = True, use_fatrop_solver = False, plot_inv = True, traj_type = "position"):
    trajectory,time = dh.read_pose_trajectory_from_txt(data_location)
    pose,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_trajectory_arclength(trajectory)
    arclength_n = arclength/arclength[-1]
    trajectory_position = pose[:,:3,3]
    trajectory_orientation = pose[:,:3,:3]

    if plot_demo:
        fig = plt.figure(figsize=(8,8))
        ax = fig.add_subplot(111, projection='3d')
        ax = plt.axes(projection='3d')
        ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'.-')
        n_frames = 10
        indx = np.trunc(np.linspace(0,len(trajectory_orientation)-1,n_frames))
        indx = indx.astype(int)
        for i in indx:
            pl.plot_3d_frame(trajectory_position[i,:],trajectory_orientation[i,:,:],1,0.01,['red','green','blue'],ax)

    #%%
    optim_calc_results = OCP_results(FSt_frames = [], FSr_frames = [], Obj_pos = [], Obj_frames = [], invariants = np.zeros((len(trajectory),6)))

    
    FS_calculation_problem_rot = FS_calc_rot(window_len=nb_samples, bool_unsigned_invariants = False, rms_error_traj = 2*pi/180, fatrop_solver = use_fatrop_solver) 
    FS_calculation_problem_pos = FS_calc_pos(window_len=nb_samples, bool_unsigned_invariants = False, rms_error_traj = 0.003, fatrop_solver = use_fatrop_solver)
    
    # specify optimization problem symbolically + calculate invariants given measurements
    if not traj_type == "position":
        if traj_type == "pose":
             
             optim_calc_results.invariants[:,3:], optim_calc_results.Obj_pos, optim_calc_results.FSt_frames = FS_calculation_problem_pos.calculate_invariants(trajectory,stepsize)
        else:
            optim_calc_results.Obj_pos = trajectory_position
        
        optim_calc_results.invariants[:,:3], optim_calc_results.Obj_frames, optim_calc_results.FSr_frames = FS_calculation_problem_rot.calculate_invariants(trajectory,stepsize)
    else:
        optim_calc_results.invariants, optim_calc_results.Obj_pos, optim_calc_results.FSt_frames = FS_calculation_problem_pos.calculate_invariants(trajectory,stepsize)
        optim_calc_results.Obj_frames = trajectory_orientation

    if plot_inv:
        fig = plt.figure(figsize=(8,8))
        ax = fig.add_subplot(111, projection='3d')
        ax = plt.axes(projection='3d')
        ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'b')
        ax.plot(optim_calc_results.Obj_pos[:,0],optim_calc_results.Obj_pos[:,1],optim_calc_results.Obj_pos[:,2],'r')
        indx = np.trunc(np.linspace(0,len(optim_calc_results.Obj_pos)-1,n_frames))
        indx = indx.astype(int)
        if not traj_type == "position":
            for i in indx:
                pl.plot_3d_frame(optim_calc_results.Obj_pos[i,:],optim_calc_results.Obj_frames[i,:,:],1,0.01,['red','green','blue'],ax)

            pl.plot_orientation(optim_calc_results.Obj_frames,trajectory_orientation)
        if traj_type == "pose":
            pl.plot_invariants(optim_calc_results.invariants,[],arclength_n)
        elif traj_type == "position":
            pl.plot_invariants(optim_calc_results.invariants,[],arclength_n, inv_type = "FS_pos")
        else:
            pl.plot_invariants(optim_calc_results.invariants,[],arclength_n, inv_type = "FS_rot")

    if plt.get_backend() != 'agg':
        plt.show()

    return optim_calc_results

#%%
def generate_trajectory(data_location, optim_calc_results, p_obj_end, rotate, use_fatrop_solver = False, plot_new_trajectory = True, traj_type = "position"):
    trajectory,time = dh.read_pose_trajectory_from_txt(data_location)
    pose,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_trajectory_arclength(trajectory)
    arclength_n = arclength/arclength[-1]
    n_frames = 10
    # Spline of model
    knots = np.concatenate(([arclength_n[0]],[arclength_n[0]],arclength_n,[arclength_n[-1]],[arclength_n[-1]]))
    degree = 3
    spline_model_trajectory = ip.BSpline(knots,optim_calc_results.invariants,degree)

    def interpolate_model_invariants(demo_invariants, progress_values):
        
        resampled_invariants = np.array([demo_invariants(i) for i in progress_values]) 
        new_stepsize = progress_values[1] - progress_values[0] 
        
        resampled_invariants[:,0] = resampled_invariants[:,0] *  (progress_values[-1] - progress_values[0])
        return resampled_invariants, new_stepsize

    #%% 
    current_progress = 0
    number_samples = 100

    progress_values = np.linspace(current_progress, arclength_n[-1], number_samples)
    model_invariants,new_stepsize = interpolate_model_invariants(spline_model_trajectory,progress_values)

    # new constraints
    current_index = round(current_progress*len(trajectory))
    p_obj_start = optim_calc_results.Obj_pos[current_index]
    R_obj_start = orthonormalize(optim_calc_results.Obj_frames[current_index])
    # alpha = 180
    # rotate = R.from_euler('z', alpha, degrees=True)
    R_obj_end =  orthonormalize(rotate.as_matrix() @ optim_calc_results.Obj_frames[-1])
    if not traj_type == "rotation":
        FSt_start = orthonormalize(optim_calc_results.FSt_frames[current_index])
        FSt_end = orthonormalize(rotate.as_matrix() @ optim_calc_results.FSt_frames[-1])

    # define new class for OCP results
    optim_gen_results = OCP_results(FSt_frames = [], FSr_frames = [], Obj_pos = [], Obj_frames = [], invariants = np.zeros((number_samples,6)))



    # Linear initialization
    R_obj_init = interpR(np.linspace(0, 1, len(optim_calc_results.Obj_frames)), [0,1], np.array([R_obj_start, R_obj_end]))
    
    if not traj_type == "position":
        R_r_init, R_r_init_array, invars_init = FSr_init(R_obj_start, R_obj_end)

    boundary_constraints = {
    "position": {
        "initial": p_obj_start,
        "final": p_obj_end
    },
    "moving-frame": {
        "initial": FSt_start,
        "final": FSt_end
    },
    }
    initial_values = {
        "trajectory": optim_calc_results.Obj_pos,
        "moving-frames": optim_calc_results.FSt_frames,
        "invariants": model_invariants[:,3:],
        "invariants-orientation": invars_init,
        "trajectory-orientation": R_obj_init,
        "moving-frame-orientation": R_r_init_array
    }

    boundary_constraints_rot = {"orientation": {"final": R_obj_end}}

    
    # specify optimization problem symbolically
    FS_online_generation_problem_pos = FS_gen_pos(boundary_constraints, window_len=number_samples, fatrop_solver = use_fatrop_solver)
    FS_online_generation_problem_rot = FS_gen_rot(boundary_constraints_rot, number_samples, fatrop_solver = use_fatrop_solver)
    
    # Define OCP weights
    w_invars_pos = np.array([5*10**1, 1.0, 1.0])
    weight_params = {"w_invars": 10**2*np.array([10**1, 1.0, 1.0])}
    weights = {}
    weights['w_invars'] = w_invars_pos

    # Solve
    if not traj_type == "position":
        optim_gen_results.invariants[:,:3], optim_gen_results.Obj_frames, optim_gen_results.FSr_frames, tot_time_rot = FS_online_generation_problem_rot.generate_trajectory(model_invariants[:,:3],boundary_constraints_rot,new_stepsize,weight_params,initial_values)
        if traj_type == "pose":
            optim_gen_results.invariants[:,3:], optim_gen_results.Obj_pos, optim_gen_results.FSt_frames, tot_time_pos = FS_online_generation_problem_pos.generate_trajectory(invariant_model = model_invariants[:,3:], initial_values=initial_values, boundary_constraints=boundary_constraints, step_size = new_stepsize, weights_params = weights)
        else:
            optim_gen_results.Obj_pos = pose[:,:3,3]
            tot_time_pos = 0
    else:
        optim_gen_results.invariants, optim_gen_results.Obj_pos, optim_gen_results.FSt_frames, tot_time_pos = FS_online_generation_problem_pos.generate_trajectory(invariant_model = model_invariants, initial_values=initial_values, boundary_constraints=boundary_constraints, step_size = new_stepsize, weights_params = weights)
        optim_gen_results.Obj_frames = pose[:,:3,:3]
        tot_time_rot = 0
    print('')
    print("TOTAL time to generate new trajectory: ")
    print(str(tot_time_pos + tot_time_rot) + "[s]")

    for i in range(len(optim_gen_results.Obj_frames)):
        optim_gen_results.Obj_frames[i] = orthonormalize(optim_gen_results.Obj_frames[i])

    if plot_new_trajectory:
        fig = plt.figure(figsize=(14,8))
        ax = fig.add_subplot(111, projection='3d')
        ax.plot(optim_calc_results.Obj_pos[:,0],optim_calc_results.Obj_pos[:,1],optim_calc_results.Obj_pos[:,2],'b')
        ax.plot(optim_gen_results.Obj_pos[:,0],optim_gen_results.Obj_pos[:,1],optim_gen_results.Obj_pos[:,2],'r')

        indx_online = np.trunc(np.linspace(0,len(optim_gen_results.Obj_pos)-1,n_frames))
        indx_online = indx_online.astype(int)
        if not traj_type == "position":
            for i in indx_online:
                pl.plot_3d_frame(optim_calc_results.Obj_pos[i,:],optim_calc_results.Obj_frames[i,:,:],1,0.01,['red','green','blue'],ax)
                pl.plot_3d_frame(optim_gen_results.Obj_pos[i,:],optim_gen_results.Obj_frames[i,:,:],1,0.01,['red','green','blue'],ax)
            pl.plot_orientation(optim_calc_results.Obj_frames,optim_gen_results.Obj_frames,current_index)

        if traj_type == "pose":
            pl.plot_invariants(optim_calc_results.invariants, optim_gen_results.invariants, arclength_n, progress_values)
        elif traj_type == "position":
            pl.plot_invariants(optim_calc_results.invariants, optim_gen_results.invariants, arclength_n, progress_values, 'FS_pos')
        else:
            pl.plot_invariants(optim_calc_results.invariants, optim_gen_results.invariants, arclength_n, progress_values, 'FS_rot')

    if plt.get_backend() != 'agg':
        plt.show()


import numpy as np
import casadi as cas
import invariants_py.dynamics_invariants as dynamics
from invariants_py import ocp_helper
from invariants_py.initialization import generate_initvals_from_bounds
from invariants_py import spline_handler as sh

class OCP_gen_pos:

    def __init__(self, N = 40, w_invars = (10**-3)*np.array([1.0, 1.0, 1.0])):  
        #%% Create decision variables and parameters for the optimization problem
        opti = cas.Opti() # use OptiStack package from Casadi for easy bookkeeping of variables 

        # Define system states X (unknown object pose + moving frame pose at every time step) 
        p_obj = [opti.variable(3,1) for _ in range(N)] # object position
        R_t = [opti.variable(3,3) for _ in range(N)] # translational Frenet-Serret frame
        X = [cas.vertcat(cas.vec(R_t[k]), cas.vec(p_obj[k])) for k in range(N)]

        # Define system controls (invariants at every time step)
        U = opti.variable(3,N-1) # invariants

        # Define system parameters P (known values in optimization that need to be set right before solving)
        h = opti.parameter(1,1) # step size for integration of dynamic model
        U_demo = opti.parameter(3,N-1) # model invariants

        # Boundary values
        R_t_start = opti.parameter(3,3)
        R_t_end = opti.parameter(3,3)
        p_obj_start = opti.parameter(3,1)
        p_obj_end = opti.parameter(3,1)

        #%% Specifying the constraints
        
        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        opti.subject_to(ocp_helper.tril_vec(R_t[0].T @ R_t[0] - np.eye(3)) == 0)
        
        # Boundary constraints
        #opti.subject_to(R_t[0] == R_t_start)
        #opti.subject_to(R_t[-1] == R_t_end)
        opti.subject_to(p_obj[0] == p_obj_start)
        opti.subject_to(p_obj[-1] == p_obj_end)
            
        # Dynamic constraints
        integrator = dynamics.define_geom_integrator_tra_FSI_casadi(h)
        for k in range(N-1):
            # Integrate current state to obtain next state (next rotation and position)
            Xk_end = integrator(X[k],U[:,k],h)
            
            # Gap closing constraint
            opti.subject_to(Xk_end==X[k+1])
            
        #%% Specifying the objective

        # Fitting constraint to remain close to measurements
        objective_fit = 0
        for k in range(N-1):
            err_invars = w_invars*(U[:,k] - U_demo[:,k])
            objective_fit = objective_fit + 1/N*cas.dot(err_invars,err_invars)
        objective = objective_fit

        #%% Define solver and save variables
        opti.minimize(objective)
        opti.solver('ipopt',{"print_time":True},{'print_level':5,'max_iter':100})
        #opti.solver('ipopt',{"print_time":True,"expand":True},{'gamma_theta':1e-12,'tol':1e-6,'print_level':5,'ma57_automatic_scaling':'no','linear_solver':'mumps','max_iter':100})
        
        # Save variables
        self.R_t = R_t
        self.p_obj = p_obj
        self.U = U
        self.U_demo = U_demo
        self.R_t_start = R_t_start
        self.R_t_end = R_t_end
        self.p_obj_start = p_obj_start
        self.p_obj_end = p_obj_end
        self.h = h
        self.N = N
        self.opti = opti
        
         
    def generate_trajectory(self,U_demo,p_obj_init,R_t_init,R_t_start,R_t_end,p_obj_start,p_obj_end,step_size):
        #%%

        N = self.N
        
        # Initialize states
        for k in range(N):
            self.opti.set_initial(self.R_t[k], R_t_init[k])
            self.opti.set_initial(self.p_obj[k], p_obj_init[k])
            
        # Initialize controls
        for k in range(N-1):    
            self.opti.set_initial(self.U[:,k], U_demo[k,:])

        # Set values boundary constraints
        self.opti.set_value(self.R_t_start,R_t_start)
        self.opti.set_value(self.R_t_end,R_t_end)
        self.opti.set_value(self.p_obj_start,p_obj_start)
        self.opti.set_value(self.p_obj_end,p_obj_end)
                
        # Set values parameters
        self.opti.set_value(self.h,step_size)
        for k in range(N-1):
            self.opti.set_value(self.U_demo[:,k], U_demo[k,:])     
        
        # ######################
        # ##  DEBUGGING: check integrator in initial values, time step 0 to 1
        # x0 = cas.vertcat(cas.vec(np.eye(3,3)), cas.vec(measured_positions[0]))
        # u0 = 1e-8*np.ones((3,1))
        # integrator = dynamics.define_geom_integrator_tra_FSI_casadi(self.stepsize)
        # x1 = integrator(x0,u0)
        # print(x1)
        # ######################

        # Solve the NLP
        sol = self.opti.solve_limited()
        self.sol = sol
        
        # Extract the solved variables
        invariants = sol.value(self.U).T
        invariants =  np.vstack((invariants,[invariants[-1,:]]))
        calculated_trajectory = np.array([sol.value(i) for i in self.p_obj])
        calculated_movingframe = np.array([sol.value(i) for i in self.R_t])
        
        return invariants, calculated_trajectory, calculated_movingframe
    
    def generate_trajectory_global(self,U_demo,initial_values,boundary_constraints,step_size):
        #%%
        N = self.N
        
        # Initialize states
        for k in range(N):
            self.opti.set_initial(self.R_t[k], initial_values["moving-frames"][k])
            self.opti.set_initial(self.p_obj[k], initial_values["trajectory"][k])
            
        # Initialize controls
        for k in range(N-1):    
            self.opti.set_initial(self.U[:,k], U_demo[k,:])

        # Set values boundary constraints
        #self.opti.set_value(self.R_t_start,boundary_constraints["moving-frames"]["initial"])
        #self.opti.set_value(self.R_t_end,boundary_constraints["moving-frames"]["final"])
        self.opti.set_value(self.p_obj_start,boundary_constraints["position"]["initial"])
        self.opti.set_value(self.p_obj_end,boundary_constraints["position"]["final"])
                
        # Set values parameters
        self.opti.set_value(self.h,step_size)
        for k in range(N-1):
            self.opti.set_value(self.U_demo[:,k], U_demo[k,:])     
        
        # Solve the NLP
        sol = self.opti.solve_limited()
        self.sol = sol
        
        # Extract the solved variables
        invariants = sol.value(self.U).T
        invariants =  np.vstack((invariants,[invariants[-1,:]]))
        calculated_trajectory = np.array([sol.value(i) for i in self.p_obj])
        calculated_movingframe = np.array([sol.value(i) for i in self.R_t])
        
        return invariants, calculated_trajectory, calculated_movingframe

def generate_trajectory_translation(invariant_model, boundary_constraints, N=40):
    
    # Specify optimization problem symbolically
    OCP = OCP_gen_pos(N = N)

    # Initial values
    initial_values, initial_values_dict = generate_initvals_from_bounds(boundary_constraints, N)

    # Resample model invariants to desired number of N samples
    spline_invariant_model = sh.create_spline_model(invariant_model[:,0], invariant_model[:,1:])
    progress_values = np.linspace(invariant_model[0,0],invariant_model[-1,0],N)
    model_invariants,progress_step = sh.interpolate_invariants(spline_invariant_model, progress_values)
    
    # Calculate remaining trajectory
    invariants, trajectory, mf = OCP.generate_trajectory_global(model_invariants,initial_values_dict,boundary_constraints,progress_step)

    return invariants, trajectory, mf, progress_values




r"""
The SE3 module provides operations on objects relevant in SE3, such as
homogeneous transformation matrices and screws. 

* **SE3/se3** : homogenous transformations in the 3-dof Euclidean space (4x4 homegeneous
  transformation matrix) / twists in the plane (6-vector consisting of 3-dof rotational
  velocity and a 3-dof velocity) 

Operations provides are conversion from other representations,
exponential/logarithms and transformation. Specialized routines for exponential
and logarithms in SE3 are provided that deliver better performance and accuracy
than the general purpose routines in e.g. SciPy. 

*Erwin Aertbelien, 2021*
"""


import numpy as np
from numpy.linalg import norm
from numpy import trace
import math
from math import fabs,copysign,sqrt,atan2,asin
from invariants_py import SO3

def random():
    """
    random_frame returns a random 4x4 homogeneous transformation matrix
    """
    [U,S,V] = np.linalg.svd(np.random.rand(3,3));
    if np.linalg.det(U) < 0:
        U[:,2]=-U[:,2];
    Trand = np.eye(4);
    Trand[:3,3 ] = np.random.rand(3);
    Trand[:3,:3] = U;
    return Trand

def inv(T):
    Ti=np.zeros((4,4))
    Ti[:3,:3] = T[:3,:3].transpose()
    Ti[:3,3]  = - Ti[:3,:3] @ T[:3,3]
    Ti[3,3] = 1
    return Ti

def frame( rot=None, p=None):
    r"""
    Creates a 4x4 homogeneous transformation matrix (in SE3)

    Parameters
    ----------
    rot : a numpy 3x3 array representing the orientation matrix
    p   : a numpy 3x1 array representing the origin of the homegeneous transformation matrix

    Returns
    -------
    a 4x4 homegeneous transformation matrix

    """
    F = np.eye(4,4)
    if not( rot is None):
        F[:3,:3] = rot
    if not( p is None):
        F[:3,3] = p
    return F

def orient(F):
    r"""
    orient(F) returns the rotation matrix part of a homegenous tf F.
    """
    return F[:3,:3]

def origin(F):
    r"""
    origin(F) returns the origin of a homegenous transformation matrix F.
    """
    return F[:3,3]

def screw_transform(T):
    r"""
    screw_transform(T) and screw_transform_se3(T) return a screw transformation
    matrix corresponding to the homegeneous transformation matrix T in se3.

    """
    S = np.zeros((6,6))
    S[0:3,0:3] = T[0:3,0:3]
    S[3:6,3:6] = T[0:3,0:3]
    S[3:6,0:3] = SO3.crossmat(T[0:3,3]) @ T[0:3,0:3]
    return S

def screw_transform_se2(T):
    r"""
    screw_transform_se2(T) returns a screw transformation
    matrix corresponding to the homegeneous transformation matrix T in se2.

    """
    S = np.eye(3)
    S[:2,:2] = T[:2,:2]
    S[2:,:2] = np.array( [-T[1,2], T[0,2]] ) @ T[:2,:2] 
    return S

def screw_orient_transform(T):
    r"""
    screw_orient_transform returns a screw transformation in se3
    matrix corresponding to the rotation matrix T.

    """
 
    S = np.zeros((6,6))
    S[0:3,0:3] = T[0:3,0:3]
    S[3:6,3:6] = T[0:3,0:3]
    return S

def screw_pos_transform(p):
    r"""
    screw_pos_transform returns a screw transformation in se3
    matrix corresponding to the position vector p.
    """

    S = np.eye(6)
    S[3:6,0:3] = SO3.crossmat(p)
    return S

def crossmat(v):
    r"""
    crossmat_se3  return a 4x4 matrix corresponding the 6x1 screw vector v  
    
    (3 dof rotation + 3 dof translation) in se3.
    (also called the "hat"-operator in se3).

    """ 
    return np.array(
        [[0, -v[2], v[1], v[3]],
         [v[2], 0, -v[0], v[4]],
         [-v[1],v[0],0  , v[5]],
         [0.,0.,0.,0.]]
    )

def crossmat_spatial(v):
    r"""
    returns Featherstone's spatial cross matrix corresponding to a screw in se3.

    Parameters
    ----------
    v : a (6,) numpy array 
        representing screw (first rotational part, then translational part)

    Returns
    -------
        a (6,6) numpy array 
            matrix containing spatial cross-product matrix:

    .. math::
    
        \begin{bmatrix}
            \omega\times  &  0            \\
            p\times       &  \omega\times
        \end{bmatrix}

    """
    OMEGA = SO3.crossmat(v[:3])
    VEL   = SO3.crossmat(v[3:])
    ZERO  = np.zeros((3,3))
    return np.block( [[OMEGA, ZERO],[VEL, OMEGA]])

def crossvec(M):
    r"""
    Returns the vector corresponding to a matrix in se3
    (also sometimes called the "vee" operator)
    (inverse of crossmat_se3)    
    
    Parameters
    ----------
    M : a 4x4 numpy array 
        Representing a matrix 

        .. math::

            \begin{bmatrix}
                \omega\times &    p \\
                0            &    0
            \end{bmatrix}

    Returns
    -------
    v : a (6,) numpy array
        vector corresponding to this input matrix

    """    
    result = np.zeros(6)
    result[:3] = SO3.crossvec(M[:3,:3])
    result[3:] = M[:3,3]
    return result

def expm(M):
    r"""
    Matrix exponential in se3 of a 4x4  matrix

    Parameters
    ----------
    M : a (4,4) numpy array
        A matrix corresponding to a displacement of the form:

        .. math::

            \begin{bmatrix}
                \omega\times & v \\
                0 & 0
            \end{bmatrix} \Delta t

    Returns
    -------
    R : a (4,4) numpy array 
        corresponding the resulting homogeneous transformation
        matrix in SE3.
 
    """
    xi = crossvec(M)
    R  = SO3.expm(M[:3,:3])
    theta = norm(xi[:3])
    result = np.eye(4)
    if theta==0:
        result[:3,3] = M[:3,3]
        return result
    v     = xi[3:] / theta
    omega = xi[:3] / theta
    G = (np.eye(3)-R) @ SO3.crossmat(omega) + np.outer(omega,omega) * theta
    result[:3,:3] = R
    result[:3,3]  = G @ v
    return result

def logm(T):
    r"""
    Matrix logarithm of a homogeneous transformation matrix in SE3

    Parameters
    ----------
    T : a (4,4) numpy array
        homogeneous transformation matrix in SE3 

    Returns
    -------
    A matrix logarithm corresponding to the displacement twist.   
    """
 
    R=T[:3,:3]
    p=T[:3,3]
    omega_hat = SO3.logm(R)
    omega     = SO3.crossvec(omega_hat)
    theta     = norm(omega)
    if math.fabs( theta )<1E-15:
        result = np.zeros((4,4))
        result[:3,3] = p 
        return result
    G = (np.eye(3)-R) @ omega_hat/theta + np.outer(omega,omega)/theta
    result = np.zeros((4,4))
    result[:3,:3] = omega_hat
    result[:3,3]  = np.linalg.inv(G) @ p *theta
    return result

def rotate_x(alpha):
    """returns a homegeneous transformation that rotates around x with alpha"""
    return frame(SO3.rot([1,0,0],alpha))

def rotate_y(alpha):
    """returns a homegeneous transformation that rotates around y with alpha"""
    return frame(SO3.rot([0,1,0],alpha))

def rotate_z(alpha):
    """returns a homegeneous transformation that rotates around z with alpha"""
    return frame(SO3.rot([0,0,1],alpha))

def rotate(axis,alpha):
    """returns a homegeneous transformation that rotates around the given axis with alpha"""
    axis = axis/norm(axis)
    return frame(SO3.rot(axis,alpha))

def translate_x(alpha):
    """returns a homegeneous transformation that translates along x with alpha"""
    return frame(p=[alpha,0,0])

def translate_y(alpha):
    """returns a homegeneous transformation that translates along y with alpha"""
    return frame(p=[0,alpha,0])

def translate_z(alpha):
    """returns a homegeneous transformation that translates along z with alpha"""
    return frame(p=[0,0,alpha])

def translate(p):
    """returns a homegeneous transformation that translates a"""
    return frame(p=p)






# TODO separate dynamic plots with Qt5 from regular plots in separate file

import numpy as np
from mpl_toolkits import mplot3d
from stl import mesh
from invariants_py.rot2quat import rot2quat
from scipy import interpolate as ip
#import PyQt5
#sys.modules.get("PyQt5")
#sys.modules.get("PyQt5.QtCore")
#import PyQt5
#import matplotlib.backends.backend_qtagg
#matplotlib.use('qt')
import matplotlib.pyplot as plt
#from mpl_toolkits import mplot3d
#import seaborn as sns
#sns.set(style='whitegrid',context='paper')
#from IPython import get_ipython

def plot_trajectory_and_bounds(boundary_constraints, trajectory):
    # Extract x, y, and z coordinates from the trajectory
    x = trajectory[:, 0]
    y = trajectory[:, 1]
    z = trajectory[:, 2]

    # Create a 3D plot
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.plot(x, y, z)

    # Plot the boundary constraints as red dots
    initial_point = boundary_constraints["position"]["initial"]
    final_point = boundary_constraints["position"]["final"]
    ax.scatter(initial_point[0], initial_point[1], initial_point[2], color='red', label='Initial Point')
    ax.scatter(final_point[0], final_point[1], final_point[2], color='red', label='Final Point')

    # Set labels for the axes
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')

    # Show the plot
    if plt.get_backend() != 'agg':
        plt.show()

def plot_trajectory(trajectory):
    # Extract x, y, and z coordinates from the trajectory
    x = trajectory[:, 0]
    y = trajectory[:, 1]
    z = trajectory[:, 2]

    # Create a 3D plot
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.plot(x, y, z)

    # Set labels for the axes
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')

    # Show the plot
    if plt.get_backend() != 'agg':
        plt.show()

def plot_trajectory_test(trajectory):
    fig = plt.figure(figsize=(8,8))
    ax = fig.add_subplot(111, projection='3d')
    ax = plt.axes(projection='3d')
    ax.view_init(elev=26, azim=140)
    ax.plot(trajectory[:,0],trajectory[:,1],trajectory[:,2],'.-')


def plotPose(pose, figure = '', label = '', c='b', m='.', orientation = False):
    """
    plots the given trajectory, positions and rotations are indicated
    @param: trajectory = list of pose matrices
    """

    if figure == '':
        fig = plt.figure()
    else:
        fig = figure
    ax = fig.gca(projection='3d')

    arrow_length = 0.01

    xq = pose[0,3]
    yq = pose[1,3]
    zq = pose[2,3]

    if orientation:
        #plot RED arrows at START
        ux = pose[0,0]
        uy = pose[1,0]
        uz = pose[2,0]

        vx = pose[0,1]
        vy = pose[1,1]
        vz = pose[2,1]

        wx = pose[0,2]
        wy = pose[1,2]
        wz = pose[2,2]

        # Make the direction data for the arrows
        a1 = ax.quiver(xq, yq, zq, ux, uy, uz, length=arrow_length, normalize=True, color = c)
        a2 = ax.quiver(xq, yq, zq, vx, vy, vz, length=arrow_length, normalize=True, color = c)
        a3 = ax.quiver(xq, yq, zq, wx, wy, wz, length=arrow_length, normalize=True, color = c)

    #plots the pose
    a4 = ax.scatter(xq, yq, zq, c=c, marker=m, label=label)

    if orientation:
        p = [a1,a2,a3,a4] #save the axis so that this pose can be removed again!!
    else:
        p = [a4]

    #labels, etc
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
#    ax.view_init(45, 45)
    ax.legend()

#    plt.autoscale(True, 'both', True)
#    plt.ion()
    if plt.get_backend() != 'agg':
        plt.show()

    return fig, p



def plotTrajectory(trajectory, figure = None, label = "trajectory", title = '', c = 'b', m ='', mark = False):
    """
    plots the given trajectory, positions and rotations are indicated
    @param: trajectory = list of pose matrices
    """
    plt.ion()
    if figure == None:
        fig = plt.figure( num = title ,figsize=(15, 13), dpi=120, facecolor='w', edgecolor='k')
    else:
        fig = figure
    ax = fig.gca(projection='3d')
#    ax = plt.axes(projection='3d')
    x = []
    y = []
    z = []

    limit_value = 0.0

#    arrow_length = 0.1

    p_lst = []
    for i in range(len(trajectory)):
        xq = trajectory[i][0,3]
        yq = trajectory[i][1,3]
        zq = trajectory[i][2,3]

        x.append(xq)
        y.append(yq)
        z.append(zq)

        if (np.abs(min([limit_value, xq,yq,zq])) > max([limit_value, xq,yq,zq])):
            limit_value = min([limit_value, xq,yq,zq])
        else:
            limit_value = max([limit_value, xq,yq,zq])



        if ((i == len(trajectory)/4) or (i == len(trajectory)/2) or (i == 3*len(trajectory)/4)) and mark:
            fig, p = plotPose(trajectory[i], fig, orientation=True, c=c)

        elif (i== 0) and mark:
            fig, p = plotPose(trajectory[i], fig, c= 'g', m = 'x', orientation=True, label='Start pose')

        elif (i == len(trajectory) -1)and mark:
            fig, p = plotPose(trajectory[i], fig, c= 'r', m = 'x', orientation=True, label='End pose')

        else:
            fig, p = plotPose(trajectory[i], fig, c = c)

        p_lst.append(p)
    #plots the line
    a1, = ax.plot(x, y, z, label = label, c = c, linestyle=m)

    #for removing every second tick if the graphs are too busy
#    for label in ax.xaxis.get_ticklabels()[::2]:
#        label.set_visible(False)
#    for label in ax.yaxis.get_ticklabels()[::2]:
#        label.set_visible(False)
#    for label in ax.zaxis.get_ticklabels()[::2]:
#        label.set_visible(False)
    p_lst.append([a1])

#    #plots the points
#    ax.scatter(x, y, z, c='k', marker='.', label = 'datapoints')
#    ax.scatter(x[0], y[0], z[0], c='r', marker='x', label = 'start')
#    ax.scatter(x[-1], y[-1], z[-1], c='g', marker='x', label = 'end')

    #labels, etc
#    ax.set_xlim(-limit_value, limit_value)
#    ax.set_ylim(-limit_value, limit_value)
#    ax.set_zlim(-limit_value, limit_value)
    ax.zaxis.set_rotate_label(False)  # disable automatic rotation
    ax.set_xlabel('$X$ [m]')
    ax.set_ylabel('$Y$ [m]')
    ax.set_zlabel('$Z$ [m]', rotation = 90)

    ax.view_init(13,-170)
#    ax.view_init(-135,135)
    ax.axis('equal')

    plt.legend(loc=2, prop={'size': 15})


#    plt.title(title)
    if plt.get_backend() != 'agg':
        plt.show()

    return fig, p_lst

def plotInvariantSignature(invariantSignature, title = 'Invariant Signature'):
    plt.figure( num = title ,figsize=(15, 13), dpi=120, facecolor='w', edgecolor='k')
    N = len(invariantSignature['U1']) + 1
    x_temp = np.linspace(0,1,N-1)
    plt.subplot(231)
    plt.plot(x_temp,[round(elem,6) for elem in invariantSignature['U1']],'r-')
    plt.title("U1 - Ir1 [rad]")
    plt.subplot(232)
    plt.plot(x_temp,invariantSignature['U2'],'r-')
    plt.title("U2 - Ir2 [rad]")
    plt.subplot(233)
    plt.plot(x_temp,invariantSignature['U3'],'r-')
    plt.title("U3 - Ir3 [rad]")
    plt.subplot(234)
    plt.plot(x_temp,[round(elem,6) for elem in invariantSignature['U4']],'r-')
    plt.title("U4 - It1 [m]")
    plt.subplot(235)
    plt.plot(x_temp,invariantSignature['U5'],'r-')
    plt.title("U5 - It2 [rad]")
    plt.subplot(236)
    plt.plot(x_temp,invariantSignature['U6'],'r-')
    plt.title("U6 - It3 [rad]")

def removeAxis(ax):
    ax.remove()
    
def plot_trajectory_invariants(trajectory,trajectory_recon,arclength_n,invariants):
    
    plt.figure(figsize=(14,6))
    plt.subplot(2,2,1)
    plt.plot(trajectory[:,0],trajectory[:,1],'.-')
    plt.plot(trajectory_recon[:,0],trajectory_recon[:,1],'.-')
    plt.title('Trajectory')
    
    plt.subplot(2,2,3)
    plt.plot(arclength_n,invariants[:,0])
    plt.plot(0,0)
    plt.title('Velocity [m/-]')
    
    plt.subplot(2,2,2)
    plt.plot(arclength_n,invariants[:,1])
    plt.plot(0,0)
    plt.title('Curvature [rad/-]')
    
    plt.subplot(2,2,4)
    plt.plot(arclength_n,invariants[:,2])
    plt.plot(0,1)
    plt.title('Torsion [rad/-]')
    
    if plt.get_backend() != 'agg':
        plt.show()

def removeMultipleAxis(pList):
    for i in range(len(pList)):
        if len(pList[i]) > 1:
            for ax in pList[i]:
                removeAxis(ax)
        else:
            removeAxis(pList[i][0])

def plot_2D_contour(trajectory):
    plt.figure(figsize=(8,3))
    plt.axis('equal')
    plt.plot(trajectory[:,0],trajectory[:,1],'.-')


def plot_trajectory_invariants_online(arclength_n, invariants, progress_values, new_invars,fig):
    
    plt.clf()

    plt.subplot(1,3,1)
    plt.plot(progress_values,new_invars[:,0],'r')
    plt.plot(arclength_n,invariants[:,0],'b')
    plt.plot(0,0)
    plt.title('velocity [m/m]')
    
    plt.subplot(1,3,2)
    plt.plot(progress_values,(new_invars[:,1]),'r')
    plt.plot(arclength_n,invariants[:,1],'b')
    plt.plot(0,0)
    plt.title('curvature [rad/m]')

    plt.subplot(1,3,3)
    plt.plot(progress_values,(new_invars[:,2]),'r')
    plt.plot(arclength_n,invariants[:,2],'b')
    plt.plot(0,0)
    plt.title('torsion [rad/m]')

    fig.canvas.draw()
    fig.canvas.flush_events()


def plot_interpolated_invariants(initial_invariants, interpolated_invariants, progress1, progress2, inv_type = 'eFS'):
    """
    plots invariants after interpolation

    Parameters
    ----------
    initial_invariants : numpy array (N,6) containing invariants before interpolation
    interpolated_invariants : numpy array (M,6) containing invariants after interpolation
    progress1 : numpy array (N,1) with the trajectory progress before interpolation
    progress2 : numpy array (M,1) with the trajectory progress after interpolation 
    inv_type : string defining the type of invariants to plot, possible entries 'eFS','FS_pos','FS_rot'
    
    """
    size = 100
    fig = plt.figure(figsize=(10,6))
    if inv_type == 'FS_rot' or inv_type == 'eFS':
        if inv_type == 'eFS':
            ax4 = fig.add_subplot(234)
            ax4.set_title('i_t1')
            ax4.plot(progress1,initial_invariants[:,3],'b')
            ax4.plot(progress2,interpolated_invariants[:,3],'r.')
            ax5 = fig.add_subplot(235)
            ax5.set_title('i_t2')
            ax5.plot(progress1,initial_invariants[:,4],'b')
            ax5.plot(progress2,interpolated_invariants[:,4],'r.')
            ax6 = fig.add_subplot(236)
            ax6.plot(progress1,initial_invariants[:,5],'b')
            ax6.plot(progress2,interpolated_invariants[:,5],'r.')
            ax6.set_title('i_t3')
            size = size + 100
        ax1 = fig.add_subplot(size + 31)
        ax1.set_title('i_r1')
        ax1.plot(progress1,initial_invariants[:,0],'b')
        ax1.plot(progress2,interpolated_invariants[:,0],'r.')
        ax2 = fig.add_subplot(size + 32)
        ax2.set_title('i_r2')
        ax2.plot(progress1,initial_invariants[:,1],'b')
        ax2.plot(progress2,interpolated_invariants[:,1],'r.')
        ax3 = fig.add_subplot(size + 33)
        ax3.set_title('i_r3')
        ax3.plot(progress1,initial_invariants[:,2],'b')
        ax3.plot(progress2,interpolated_invariants[:,2],'r.')
    elif inv_type == 'FS_pos':
        ax4 = fig.add_subplot(131)
        ax4.set_title('i_t1')
        ax4.plot(progress1,initial_invariants[:,0],'b')
        ax4.plot(progress2,interpolated_invariants[:,0],'r.')
        ax5 = fig.add_subplot(132)
        ax5.set_title('i_t2')
        ax5.plot(progress1,initial_invariants[:,1],'b')
        ax5.plot(progress2,interpolated_invariants[:,1],'r.')
        ax6 = fig.add_subplot(133)
        ax6.plot(progress1,initial_invariants[:,2],'b')
        ax6.plot(progress2,interpolated_invariants[:,2],'r.')
        ax6.set_title('i_t3')
        


def plot_invariants(invariants1, invariants2, progress1, progress2=[], inv_type='eFS', fig=None):
    """
    Plots invariants before and after interpolation.

    Parameters
    ----------
    invariants1 : numpy array
        Invariants before interpolation.
    invariants2 : numpy array
        Invariants after interpolation.
    progress1 : numpy array
        Trajectory progress before interpolation.
    progress2 : numpy array, optional
        Trajectory progress after interpolation. Default is an empty array.
    inv_type : str, optional
        Type of invariants to plot. Possible values are 'eFS', 'FS_pos', 'FS_rot'. Default is 'eFS'.
    fig : matplotlib Figure, optional
        Handle to an existing figure. If provided, the plot will be updated on the existing figure. Default is None.

    Returns
    -------
    None
    """
    if fig is None:
        fig = plt.figure(figsize=(10, 6))
    if inv_type == 'FS_rot' or inv_type == 'eFS':
        if inv_type == 'eFS':
            for i in range(3):
                ax = fig.add_subplot(2, 3, i+4, title=f'$i_{{t{i+1}}}$')
                ax.plot(progress1, invariants1[:, i+3], 'b')
                if len(invariants2):
                    ax.plot(progress2, invariants2[:, i+3], 'r')
        for i in range(3):
            if inv_type == 'eFS':
                ax = fig.add_subplot(2, 3, 1 + i, title=f'$i_{{r{i+1}}}$')
            else:
                ax = fig.add_subplot(1, 3, 1 + i, title=f'$i_{{r{i+1}}}$')
            ax.plot(progress1, invariants1[:, i], 'b')
            if len(invariants2):
                ax.plot(progress2, invariants2[:, i], 'r')
    elif inv_type == 'FS_pos':
        for i in range(3):
            ax = fig.add_subplot(1, 3, i+1, title=f'$i_{{t{i+1}}}$')
            ax.plot(progress1, invariants1[:, i], 'b')
            if len(invariants2):
                ax.plot(progress2, invariants2[:, i], 'r')

    fig.canvas.draw()
    fig.canvas.flush_events()

def plot_stl(stl_file_location,pos,R,colour,alpha,ax):
    """
    Insert a 3D mesh into an existing 3D plot by reading an stl file

    Parameters
    ----------
    stl_file_location: location of stl file conatining the desired mesh
    R: desired orientation of the 3D mesh
    pos: desired position of the 3D mesh
    colour: desired colour of the mesh
    ax: plot to which the mesh is added

    Returns
    -------
    a 3D plot in which the 3D mesh is inserted

    """
    stl_mesh = mesh.Mesh.from_file(stl_file_location)
    Tr  = np.vstack((np.hstack((R,np.array([pos]).T)), [0,0,0,1]))
    stl_mesh.transform(Tr)
    collection = mplot3d.art3d.Poly3DCollection(stl_mesh.vectors)
    collection.set_facecolor(colour); collection.set_alpha(alpha)
    ax.add_collection3d(collection)


def compare_invariants(invariants, new_invariants, arclength_n, progress_values):

    fig = plt.figure()
    plt.subplot(1,3,1)
    plt.plot(progress_values,new_invariants[:,0],'r')
    plt.plot(arclength_n,invariants[:,0],'b')
    plt.plot(0,0)
    plt.title('Velocity [m/m]')

    plt.subplot(1,3,2)
    plt.plot(progress_values,(new_invariants[:,1]),'r')
    plt.plot(arclength_n,invariants[:,1],'b')
    plt.plot(0,0)
    plt.title('Curvature [rad/m]')

    plt.subplot(1,3,3)
    plt.plot(progress_values,(new_invariants[:,2]),'r')
    plt.plot(arclength_n,invariants[:,2],'b')
    plt.plot(0,0)
    plt.title('Torsion [rad/m]')

    if plt.get_backend() != 'agg':
        plt.show()

def plot_orientation(trajectory1,trajectory2,current_index = 0):
    """
    plots the value of qx,qy,qz,qw of two rotational trajectories expressed as quaternions, starting from rotational matrices

    Parameters
    ----------
    trajectory1 : a (N,3,3) numpy array describing N orthonormal rotation matrices of the first trajectory
    trajectory2 : a (N,3,3) numpy array describing N orthonormal rotation matrices of the second trajectory
    current_index : index at which the generation of new trajectory starts (used for online generation, otherwise leave = 0 as default)
        
    Returns
    -------
    Figure with four subplots showing the values of qx,qy,qz,qw along the progress of two rotational trajectories 

    """
    R1 = trajectory1
    quat1 = rot2quat(R1)
    R2 = trajectory2
    quat2 = rot2quat(R2)
    online_n_samples = np.array(range(current_index,len(quat1)))
    interp = ip.interp1d(np.array(range(len(online_n_samples))),online_n_samples)
    x = interp(np.linspace(0,len(online_n_samples)-1,len(quat2)))

    plt.figure(figsize=(14,6))
    for i in range(np.size(quat1,1)):
        plt.subplot(np.size(quat1,1),1,i+1)
        plt.plot(quat1[:,i],'-b')
        plt.plot(x,quat2[:,i],'-r')
        if i == 0:
            plt.ylabel('q_x')
        elif i == 1:
            plt.ylabel('q_y')
        elif i == 2:
            plt.ylabel('q_z')
        elif i == 3:
            plt.ylabel('q_w')

    # TODO plot instantaneous rotational axis; use getRot from SO3.py
    # axang1 = quat.as_rotation_vector(R1)
    # axang2 = quat.as_rotation_vector(R2)

    # figure;
    # plot3(trajectory1.Obj_location(:,1),trajectory1.Obj_location(:,2),trajectory1.Obj_location(:,3))
    # hold on;
    # arrow3(trajectory1.Obj_location,trajectory1.Obj_location+0.05*axang1(:,1:3),['_r' 0.2],0.5,1)
    # arrow3(trajectory2.Obj_location,trajectory2.Obj_location+0.05*axang2(:,1:3),['_b' 0.2],0.5,1)
    # axis equal; grid on; box on;

def plot_invariants_new(invariants,arclength):
    plt.figure()
    plt.plot(arclength, invariants[:, 0], label='$v$ [m]', color='r')
    plt.plot(arclength, invariants[:, 1], label='$\omega_\kappa$ [rad/m]', color='g')
    plt.plot(arclength, invariants[:, 2], label='$\omega_\u03C4$ [rad/m]', color='b')
    plt.xlabel('s [m]')
    plt.legend()
    plt.title('Calculated invariants (full horizon)')
    if plt.get_backend() != 'agg':
        plt.show()
    #plt.close()

def plot_invariants_new(invariants, arclength):
    
    fig = plt.figure()
    plt.subplot(1,3,1)
    plt.plot(arclength,invariants[:,0],'b')
    plt.plot(0,0)
    plt.title('Velocity [m/m]')

    plt.subplot(1,3,2)
    plt.plot(arclength,invariants[:,1],'b')
    plt.plot(0,0)
    plt.title('Curvature [rad/m]')

    plt.subplot(1,3,3)
    plt.plot(arclength,invariants[:,2],'b')
    plt.plot(0,0)
    plt.title('Torsion [rad/m]')

    if plt.get_backend() != 'agg':
        plt.show()


def plot_3d_frame(p,R,scale_arrow,length_arrow,my_color,ax3d):
    """
    goal:
    plot a 3D coordinate frame

    input:
        p: 
        R: 
        scale_arrow: 
        length_arrow: 
        my_color: 
        ax3d: 
    """
    m = np.shape(R)[0]
    n = np.shape(R)[1]
    if m == 4 and (n == 4 or n == 3):
        R_new = R[:,0:2,0:2]
        R = R_new
    
    obj_Rx = R[:,[0]]*length_arrow
    obj_Ry = R[:,[1]]*length_arrow
    obj_Rz = R[:,[2]]*length_arrow
    
    ax3d.plot([p[0],p[0]+obj_Rx[0,0]], \
              [p[1],p[1]+obj_Rx[1,0]], \
              [p[2],p[2]+obj_Rx[2,0]], \
              color = my_color[0], alpha = 1, linewidth = scale_arrow)
    ax3d.plot([p[0],p[0]+obj_Ry[0,0]], \
              [p[1],p[1]+obj_Ry[1,0]], \
              [p[2],p[2]+obj_Ry[2,0]], \
              color = my_color[1], alpha = 1, linewidth = scale_arrow)
    ax3d.plot([p[0],p[0]+obj_Rz[0,0]], \
              [p[1],p[1]+obj_Rz[1,0]], \
              [p[2],p[2]+obj_Rz[2,0]], \
              color = my_color[2], alpha = 1, linewidth = scale_arrow)
    ax3d.set_xlabel('x [m]')
    ax3d.set_ylabel('y [m]')
    ax3d.set_zlabel('z [m]')


import numpy as np
import math as math
from mpl_toolkits import mplot3d
from stl import mesh





import numpy as np
import casadi as cas
import warnings

def jerk_invariant(i1,i1dot,i1ddot,i2,i2dot,i3):
    # This is the jerk of the trajectory expressed in terms of the invariants and their derivatives
    jerk = cas.vertcat(-i1*i2**2 + i1ddot, -i1*i2dot - 2*i2*i1dot, i1*i2*i3)
    return jerk

def tril_vec(input):
    return cas.vertcat(input[0,0], input[1,1], input[2,2], input[1,0], input[2,0], input[2,1])

def weighted_sum_of_squares(weights, var):
    return cas.dot(weights, var**2)

def tril_vec_no_diag(input):
    return cas.vertcat(input[1,0], input[2,0], input[2,1])

def three_elements(self,input):
    return cas.vertcat(input[0,0], input[1,0], input[2,1])

def diffR(self,input1,input2):
    dotproduct = cas.dot(input1[:,1],input2[:,1]) - 1
    error_x0 = input1[0,0] - input2[0,0]
    error_x1 = input1[1,0] - input2[1,0]
    return cas.vertcat(dotproduct, error_x0, error_x1)

def diag(self,input):
    return cas.vertcat(input[0,0], input[1,1], input[2,2])

def check_solver(fatrop_solver):
    try: # check if fatropy is installed, otherwise use ipopt
        import fatropy
        pass
    except:
        if fatrop_solver:
            print("")
            warnings.warn("Fatrop solver is not installed! Using ipopt solver instead...")
            fatrop_solver = False
    return fatrop_solver

def solution_check_pos(p_obj_m,p_obj,rms = 10**-2):
    N = p_obj.shape[0]
    tot_ek = 0
    for i in range(N):
        ek = cas.dot(p_obj[i] - p_obj_m[i],p_obj[i] - p_obj_m[i])
        tot_ek += ek
        if tot_ek > N*rms**2:
            print("")
            print("Value of error is" , np.sqrt(tot_ek/N), "and should be less than", rms)
            raise Exception("The constraint is not satisfied! Something is wrong in the calculation")           

def solution_check_rot(R_obj_m,R_obj,rms = 4*np.pi/180):
    N = R_obj.shape[0]
    tot_ek = 0
    for i in range(N):
        ek = cas.dot(R_obj_m[i].T @ R_obj[i] - np.eye(3),R_obj_m[i].T @ R_obj[i] - np.eye(3))
        tot_ek +=ek
        if tot_ek > N*rms**2:
            print("")
            print("Value of error is" , np.sqrt(tot_ek/N), "and should be less than", rms)
            raise Exception("The constraint is not satisfied! Something is wrong in the calculation")        



import numpy as np
import math as math
from invariants_py.orthonormalize_rotation import orthonormalize_rotation as orthonormalize

def rot2quat(R_all):
    """
    Transform a 3x3 rotational matrix into the corresponding unit quaternion

        Parameters 
        ----------
        R_all : a (3,3) numpy array describing a rotational matrix (or a series of rotational matrices)

        Returns
        ----------
        The corresposing unit quaternion of the form [qx,qy,qz,qw]

    Note: one rotation matrix may correspond to two quaternions: +q and -q.
    This ambiguity is normally solved by assuming the scalar part positive.
    However, for quaternions in a time series we desire continuity so we will 
    look at the previous time sample to determine if the sign must be positive 
    or negative for continuity
    """
    N = np.size(R_all,0)
    q_all = np.zeros((N,4))

    for i in range(N):
        R = R_all[i,:,:]
        R = orthonormalize(R)
        
        qs = np.sqrt(np.trace(R)+1)/2.0
        kx = R[2,1] - R[1,2]   # Oz - Ay
        ky = R[0,2] - R[2,0]   # Ax - Nz
        kz = R[1,0] - R[0,1]   # Ny - Ox

        if (R[0,0] >= R[1,1]) and (R[0,0] >= R[2,2]):
            kx1 = R[0,0] - R[1,1] - R[2,2] + 1 # Nx - Oy - Az + 1
            ky1 = R[1,0] + R[0,1]              # Ny + Ox
            kz1 = R[2,0] + R[0,2]              # Nz + Ax
            add = (kx >= 0)
        elif (R[1,1] >= R[2,2]):
            kx1 = R[1,0] + R[0,1]          # Ny + Ox
            ky1 = R[1,1] - R[0,0] - R[2,2] + 1 # Oy - Nx - Az + 1
            kz1 = R[2,1] + R[1,2]          # Oz + Ay
            add = (ky >= 0)
        else:
            kx1 = R[2,0] + R[0,2]          # Nz + Ax
            ky1 = R[2,1] + R[1,2]          # Oz + Ay
            kz1 = R[2,2] - R[0,0] - R[1,1] + 1 # Az - Nx - Oy + 1
            add = (kz >= 0)

        if add:
            kx = kx + kx1
            ky = ky + ky1
            kz = kz + kz1
        else:
            kx = kx - kx1
            ky = ky - ky1
            kz = kz - kz1
        nm = np.linalg.norm([kx,ky,kz])
        if nm == 0:
            q = [0,0,0,1] # notation [vector  scalar] here 
        else:
            s = np.sqrt(1 - qs**2) / nm
            qv = s*np.array([kx,ky,kz])

            q = np.hstack((qv,qs)) # notation [vector  scalar] here 

            if i>1 and np.linalg.norm(q - q_all[i-1,:])>0.5:
                q = -q
        
        q_all[i,:] = q
        
    return q_all



# Imports
import numpy as np
from math import pi
import invariants_py.data_handler as dh
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import invariants_py.reparameterization as reparam
import scipy.interpolate as ip
from invariants_py.opti_calculate_vector_invariants_rotation import OCP_calc_rot
from invariants_py.opti_generate_rotation_from_vector_invariants import OCP_gen_rot
from IPython.display import clear_output
from invariants_py.plotters import plot_3d_frame, plot_orientation, plot_stl
from stl import mesh
from scipy.spatial.transform import Rotation as R
from invariants_py.orthonormalize_rotation import orthonormalize_rotation as orthonormalize
#%%
data_location = dh.find_data_path('beer_1.txt')
opener_location = dh.find_data_path('opener.stl')
trajectory,time = dh.read_pose_trajectory_from_txt(data_location)
pose,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_trajectory_arclength(trajectory)
arclength_n = arclength/arclength[-1]
trajectory_position = pose[:,:3,3]
trajectory_orientation = pose[:,:3,:3]

fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111, projection='3d')
ax = plt.axes(projection='3d')
ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'.-')
n_frames = 10
indx = np.trunc(np.linspace(0,len(trajectory_orientation)-1,n_frames))
indx = indx.astype(int)

for i in indx:
    plot_3d_frame(trajectory_position[i,:],trajectory_orientation[i,:,:],1,0.05,['red','green','blue'],ax)
    plot_stl(opener_location,trajectory_position[i,:],trajectory_orientation[i,:,:],colour="c",alpha=0.2,ax=ax)

#%%
# specify optimization problem symbolically
FS_calculation_problem = OCP_calc_rot(window_len=nb_samples, bool_unsigned_invariants = False, rms_error_traj = 2*pi/180) 

# calculate invariants given measurements
invariants, calculate_trajectory, movingframes = FS_calculation_problem.calculate_invariants_global(trajectory,stepsize)
init_vals_calculate_trajectory = calculate_trajectory
init_vals_movingframes = movingframes

fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111, projection='3d')
ax = plt.axes(projection='3d')
ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'.-')
indx = np.trunc(np.linspace(0,len(calculate_trajectory)-1,n_frames))
indx = indx.astype(int)
for i in indx:
    plot_3d_frame(trajectory_position[i,:],calculate_trajectory[i,:,:],1,0.05,['red','green','blue'],ax)
    plot_stl(opener_location,trajectory_position[i,:],calculate_trajectory[i,:,:],colour="r",alpha=0.2,ax=ax)
    plot_stl(opener_location,trajectory_position[i,:],trajectory_orientation[i,:,:],colour="c",alpha=0.2,ax=ax)

plot_orientation(init_vals_calculate_trajectory,trajectory_orientation)

f, (ax1, ax2, ax3) = plt.subplots(1, 3, sharey=True, figsize=(10,3))
ax1.plot(arclength_n,invariants[:,0])
ax1.set_title('Velocity [m/m]')
ax2.plot(arclength_n,invariants[:,1])
ax2.set_title('Curvature [rad/m]')
ax3.plot(arclength_n,invariants[:,2])
ax3.set_title('Torsion [rad/m]')

if plt.get_backend() != 'agg':
    plt.show()

#%%
# Spline of model
knots = np.concatenate(([arclength_n[0]],[arclength_n[0]],arclength_n,[arclength_n[-1]],[arclength_n[-1]]))
degree = 3
spline_model_trajectory = ip.BSpline(knots,invariants,degree)

def interpolate_model_invariants(demo_invariants, progress_values):
    
    resampled_invariants = np.array([demo_invariants(i) for i in progress_values]) 
    new_stepsize = progress_values[1] - progress_values[0] 
    
    resampled_invariants[:,0] = resampled_invariants[:,0] *  (progress_values[-1] - progress_values[0])
    return resampled_invariants, new_stepsize

#%% 
current_progress = 0.3
number_samples = 60

progress_values = np.linspace(current_progress, arclength_n[-1], number_samples)
model_invariants,new_stepsize = interpolate_model_invariants(spline_model_trajectory,progress_values)

f, (ax1, ax2, ax3) = plt.subplots(1, 3, sharey=True, figsize=(10,3))
ax1.plot(arclength_n,invariants[:,0])
ax1.plot(progress_values,model_invariants[:,0],'r.')
ax1.set_title('Velocity [m/m]')
ax2.plot(arclength_n,invariants[:,1])
ax2.plot(progress_values,model_invariants[:,1],'r.')
ax2.set_title('Curvature [rad/m]')
ax3.plot(arclength_n,invariants[:,2])
ax3.plot(progress_values,model_invariants[:,2],'r.')
ax3.set_title('Torsion [rad/m]')

# new constraints
current_index = round(current_progress*len(trajectory))
R_obj_start = orthonormalize(calculate_trajectory[current_index,:,:])
rotate = R.from_euler('z', 30, degrees=True)
R_obj_end =  orthonormalize(rotate.apply(calculate_trajectory[-1]))
R_r_start = orthonormalize(movingframes[current_index])
R_r_end = orthonormalize(movingframes[-1])


# specify optimization problem symbolically
FS_online_generation_problem = OCP_gen_rot(window_len=number_samples,w_invars = 10**2*np.array([10**1, 1.0, 1.0]))

# Solve
new_invars, new_trajectory, new_movingframes = FS_online_generation_problem.generate_trajectory(U_demo = model_invariants, R_obj_init = calculate_trajectory, R_r_init = movingframes, R_r_start = R_r_start, R_r_end = R_r_end, R_obj_start = R_obj_start, R_obj_end = R_obj_end, step_size = new_stepsize)
for i in range(len(new_trajectory)):
    new_trajectory[i] = orthonormalize(new_trajectory[i])

fig = plt.figure(figsize=(14,8))
ax = fig.add_subplot(111, projection='3d')
ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'.-')
for i in indx:
    plot_stl(opener_location,trajectory_position[i,:],calculate_trajectory[i,:,:],colour="c",alpha=0.2,ax=ax)

cut_trajectory_position = trajectory_position[current_index:,:]
interp = ip.interp1d(np.linspace(0,1,len(cut_trajectory_position)),cut_trajectory_position, axis = 0)
trajectory_position_online = interp(np.linspace(0,1,len(new_trajectory)))
indx_online = np.trunc(np.linspace(0,len(new_trajectory)-1,n_frames))
indx_online = indx_online.astype(int)
for i in indx_online:
    plot_3d_frame(trajectory_position_online[i,:],new_trajectory[i,:,:],1,0.05,['red','green','blue'],ax)
    plot_stl(opener_location,trajectory_position_online[i,:],new_trajectory[i,:,:],colour="r",alpha=0.2,ax=ax)
plot_orientation(calculate_trajectory, new_trajectory,current_index)

fig = plt.figure()
plt.subplot(1,3,1)
plt.plot(progress_values,new_invars[:,0],'r')
plt.plot(arclength_n,invariants[:,0],'b')
plt.plot(0,0)
plt.title('Velocity [m/m]')

plt.subplot(1,3,2)
plt.plot(progress_values,(new_invars[:,1]),'r')
plt.plot(arclength_n,invariants[:,1],'b')
plt.plot(0,0)
plt.title('Curvature [rad/m]')

plt.subplot(1,3,3)
plt.plot(progress_values,(new_invars[:,2]),'r')
plt.plot(arclength_n,invariants[:,2],'b')
plt.plot(0,0)
plt.title('Torsion [rad/m]')

if plt.get_backend() != 'agg':
    plt.show()

#%% Visualization

window_len = 20

# specify optimization problem symbolically
FS_online_generation_problem = OCP_gen_rot(window_len=window_len,w_invars = 10**1*np.array([10**1, 1.0, 1.0]))

current_progress = 0.0
old_progress = 0.0

R_obj_end = calculate_trajectory[-1] # initialise R_obj_end with end point of reconstructed trajectory
iterative_trajectory = calculate_trajectory.copy()
iterative_movingframes = movingframes.copy()
trajectory_position_iter = trajectory_position.copy()

fig_traj = plt.figure(figsize=(14,8))
ax = fig_traj.add_subplot(111, projection='3d')    

fig_invars, axes = plt.subplots(1, 3, sharey=True, figsize=(10,3))
    
while current_progress <= 1.0:
    
    print(f"current progress = {current_progress}")

    # Resample invariants for current progress
    progress_values = np.linspace(current_progress, arclength_n[-1], window_len)
    model_invariants,new_stepsize = interpolate_model_invariants(spline_model_trajectory,progress_values)
    
    # Boundary constraints
    current_index = round( (current_progress - old_progress) * len(iterative_trajectory))
    R_obj_start = iterative_trajectory[current_index]
    rotate = R.from_euler('z', 30/window_len, degrees=True)
    R_obj_end =  orthonormalize(rotate.apply(R_obj_end))
    R_r_start = iterative_movingframes[current_index] 
    R_r_end = iterative_movingframes[-1] 

    # Calculate remaining trajectory
    new_invars, iterative_trajectory, iterative_movingframes = FS_online_generation_problem.generate_trajectory(U_demo = model_invariants, R_obj_init = calculate_trajectory, R_r_init = movingframes, R_r_start = R_r_start, R_r_end = R_r_end, R_obj_start = R_obj_start, R_obj_end = R_obj_end, step_size = new_stepsize)
    for i in range(len(iterative_trajectory)):
        iterative_trajectory[i] = orthonormalize(iterative_trajectory[i])

    # Visualization
    clear_output(wait=True)
    
    ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'b')
    for i in indx:
        plot_stl(opener_location,trajectory_position[i,:],calculate_trajectory[i,:,:],colour="c",alpha=0.2,ax=ax)
    cut_trajectory_position = trajectory_position_iter[current_index:,:]
    interp = ip.interp1d(np.linspace(0,1,len(cut_trajectory_position)),cut_trajectory_position, axis = 0)
    trajectory_position_iter = interp(np.linspace(0,1,len(iterative_trajectory)))
    indx_iter = np.trunc(np.linspace(0,len(iterative_trajectory)-1,n_frames))
    indx_iter = indx_iter.astype(int)
    for i in indx_iter:
        plot_3d_frame(trajectory_position_iter[i,:],iterative_trajectory[i,:,:],1,0.05,['red','green','blue'],ax)
        plot_stl(opener_location,trajectory_position_iter[i,:],iterative_trajectory[i,:,:],colour="r",alpha=0.2,ax=ax)
    if plt.get_backend() != 'agg':
        plt.show()

    axes[0].plot(progress_values,new_invars[:,0],'r')
    axes[0].plot(arclength_n,invariants[:,0],'b')
    axes[0].plot(0,0)
    axes[0].set_title('velocity [m/m]')
    
    axes[1].plot(progress_values,(new_invars[:,1]),'r')
    axes[1].plot(arclength_n,invariants[:,1],'b')
    axes[1].plot(0,0)
    axes[1].set_title('curvature [rad/m]')
    
    axes[2].plot(progress_values,(new_invars[:,2]),'r')
    axes[2].plot(arclength_n,invariants[:,2],'b')
    axes[2].plot(0,0)
    axes[2].set_title('torsion [rad/m]')


    old_progress = current_progress
    current_progress = old_progress + 1/window_len


import numpy as np
from invariants_py.reparameterization import interpR

R_start = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])  # Rotation matrix 1
R_end = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]])  # Rotation matrix 2

# Interpolate between R_start and R_end
R_interp = interpR(np.linspace(0, 1, 100), [0,1], np.array([R_start, R_end]))

print(R_interp)


import rockit
import numpy as np
import casadi as cas
import invariants_py.dynamics_invariants as dynamics
import invariants_py.ocp_helper as ocp_helper
from invariants_py import initialization
from invariants_py.ocp_helper import check_solver

class OCP_calc_pos:

    def __init__(self, nb_samples = 100, w_pos = 1, w_regul_jerk = 10**-6 , w_regul_invars = 10**-10, fatrop_solver = False, planar_task = False):
        fatrop_solver = check_solver(fatrop_solver)               

        #%% Decision variables and parameters for the optimization problem 
        self.ocp = rockit.Ocp(T=1.0)        

        # States
        self.p_obj = self.ocp.state(3) # object position
        self.R_t_x = self.ocp.state(3,1) # Frenet-Serret frame, first axis
        self.R_t_y = self.ocp.state(3,1) # Frenet-Serret frame, second axis
        self.R_t_z = self.ocp.state(3,1) # Frenet-Serret frame, third axis
        R_t = cas.horzcat(self.R_t_x,self.R_t_y,self.R_t_z)
        self.i1 = self.ocp.state(1)
        self.i1dot = self.ocp.state(1)
        self.i2 = self.ocp.state(1)
        
        # Controls
        self.i1ddot = self.ocp.control(1)
        self.i2dot = self.ocp.control(1)
        self.i3 = self.ocp.control(1)
        invariants = cas.vertcat(self.i1,self.i2,self.i3)
        
        # Parameters
        self.p_obj_m = self.ocp.register_parameter(cas.MX.sym('p_obj_m',3),grid='control+')
        self.h = self.ocp.register_parameter(cas.MX.sym('step_size'))

        #%% Constraints

        # test initial constraint on position
        self.ocp.subject_to(self.ocp.at_tf(self.p_obj == self.p_obj_m))
        self.ocp.subject_to(self.ocp.at_t0(self.p_obj == self.p_obj_m))

        # Orthonormality of rotation matrix (only needed for one timestep, property is propagated by integrator)  
        self.ocp.subject_to(self.ocp.at_t0(ocp_helper.tril_vec(R_t.T @ R_t - np.eye(3))==0.))

        # Dynamics equations        
        (R_t_plus1, p_obj_plus1) = dynamics.vector_invariants_position(R_t, self.p_obj, invariants, self.h)
        
        self.ocp.set_next(self.R_t_x,R_t_plus1[:,0])
        self.ocp.set_next(self.R_t_y,R_t_plus1[:,1])
        self.ocp.set_next(self.R_t_z,R_t_plus1[:,2])
        self.ocp.set_next(self.p_obj,p_obj_plus1)
        self.ocp.set_next(self.i1,self.i1 + self.i1dot * self.h + self.i1ddot * self.h**2/2)
        self.ocp.set_next(self.i1dot,self.i1dot + self.i1ddot * self.h)
        self.ocp.set_next(self.i2,self.i2 + self.i2dot * self.h)
                     
        # Planar task constraint
        if planar_task:
            # Constraint the third axis of moving frame (binormal axis) lies along the vertical direction
            # a . b > 0 -> a and b are pointing in the same direction
            # a . b < 0 -> a and b are pointing in the opposite direction 
            self.ocp.subject_to(cas.dot( self.R_t_z , np.array([0,0,1]) ) > 0)  

        #%% Objective function
        self.N_controls = nb_samples-1

        # Term 1: Measurement fitting constraint
        obj_fit = self.ocp.sum(1/self.N_controls*w_pos*cas.dot(self.p_obj-self.p_obj_m,self.p_obj-self.p_obj_m))

        # Term 2: Regularization on jerk of trajectory
        jk = ocp_helper.jerk_invariant(self.i1,self.i1dot,self.i1ddot,self.i2,self.i2dot,self.i3)
        obj_reg_1 = self.ocp.sum(1/self.N_controls*w_regul_jerk*cas.dot(jk,jk))

        # Term 3: Regularization on moving frame invariants
        inv_controls = cas.vertcat(self.i2,self.i3)
        obj_reg_2 = self.ocp.sum(1/self.N_controls*w_regul_invars*cas.dot(inv_controls,inv_controls))

        # Sum of terms
        self.ocp.add_objective(obj_fit)
        self.ocp.add_objective(obj_reg_1)
        self.ocp.add_objective(obj_reg_2)
        
        #%% Solver options
        
        if fatrop_solver: 
            method = rockit.external_method('fatrop', N=self.N_controls)
            self.ocp.method(method)
        else:
            self.ocp.method(rockit.MultipleShooting(N=self.N_controls))
            #ocp.solver('ipopt', {'expand':True})
            silent_ipopt_options = {"print_time":False,"expand":True,"ipopt.print_level":0,'ipopt.max_iter':200,'ipopt.tol':1e-8}
            self.ocp.solver('ipopt', silent_ipopt_options)
        
        # Solve already once with dummy measurements
        self.initialize_solver(nb_samples)
        if fatrop_solver:
            self.ocp._method.set_option("print_level",0)
            self.ocp._method.set_option("tol",1e-12)
        self.first_window = True
        
        # Transform the whole OCP to a Casadi function
        self.define_ocp_to_function()
    
    def define_ocp_to_function(self):
        # Encapsulate whole rockit specification in a Casadi function
        self.ocp_to_function = self.ocp.to_function('fastsolve', 
            [ # Inputs
                self.ocp.value(self.h),
                self.ocp.sample(self.p_obj_m, grid='control')[1],
                self.ocp.sample(self.R_t_x, grid='control')[1], 
                self.ocp.sample(self.R_t_y, grid='control')[1],
                self.ocp.sample(self.R_t_z, grid='control')[1],
                self.ocp.sample(self.p_obj, grid='control')[1],
                self.ocp.sample(self.i1dot, grid='control')[1],
                self.ocp.sample(self.i1,    grid='control')[1],
                self.ocp.sample(self.i2,    grid='control')[1],
                self.ocp.sample(self.i1ddot,grid='control-')[1],
                self.ocp.sample(self.i2dot, grid='control-')[1],
                self.ocp.sample(self.i3,    grid='control-')[1], 
            ],
            [  # Outputs
                self.ocp.sample(self.R_t_x, grid='control')[1],
                self.ocp.sample(self.R_t_y, grid='control')[1],
                self.ocp.sample(self.R_t_z, grid='control')[1],
                self.ocp.sample(self.p_obj, grid='control')[1],
                self.ocp.sample(self.i1dot, grid='control')[1],
                self.ocp.sample(self.i1,    grid='control')[1],
                self.ocp.sample(self.i2,    grid='control')[1],
                self.ocp.sample(self.i1ddot,grid='control')[1],
                self.ocp.sample(self.i2dot, grid='control')[1],
                self.ocp.sample(self.i3,    grid='control')[1], 
            ],
            ["h","p_obj_m","R_t_x","R_t_y","R_t_z","p_obj","i1dot","i1","i2","i1ddot","i2dot","i3"],   # Input labels
            ["R_t_x2","R_t_y2","R_t_z2","p_obj2","i1dot2","i12","i22","i1ddot2","i2dot2","i32"],   # Output labels
        )
        
    def initialize_solver(self,window_len):
        self.ocp.set_initial(self.R_t_x, np.array([1,0,0]))                 
        self.ocp.set_initial(self.R_t_y, np.array([0,1,0]))                
        self.ocp.set_initial(self.R_t_z, np.array([0,0,1]))
        self.ocp.set_initial(self.i2,1e-12)
        self.ocp.set_value(self.p_obj_m, np.zeros((3,window_len)))
        self.ocp.set_value(self.h,1)
        self.ocp.solve_limited()
        
    def calculate_invariants_online(self,measured_positions,stepsize,window_step=1):
        #%%
        use_previous_solution = False 
        # ---> Added by Arno, it might not be the best way to always use the previous solution as initialisation for the next window. 
        # For example, if at one instance the solution does not converge, but instead diverges to a very nervous solution, 
        # this solution would then be used as an initialisation for the next window, which is not desired. 
        # We want to get rid of that 'outlier solution' as soon as possible to increase the overall robustness of this online calculation. 

        if not(use_previous_solution) or self.first_window: 
            N = self.N_controls

            Pdiff = np.diff(measured_positions, axis=0)
            Pdiff = np.vstack((Pdiff, Pdiff[-1]))
            [ex,ey,ez] = initialization.estimate_initial_frames(Pdiff)

            #[ex,ey,ez] = initialization.initialize_VI_pos(measured_positions)
            
            self.R_t_x_sol =  ex.T 
            self.R_t_y_sol =  ey.T 
            self.R_t_z_sol =  ez.T 
            self.p_obj_sol =  measured_positions.T 
            self.i1dot_sol =  1e-12*np.ones((1,N+1)) 
            self.i1_sol =     1e0*np.ones((1,N+1)) 
            self.i2_sol =     1e-1*np.ones((1,N+1)) 
            self.i1ddot_sol = 1e-12*np.ones((1,N)) 
            self.i2dot_sol =  1e-12*np.ones((1,N)) 
            self.i3_sol =     1e-12*np.ones((1,N)) 
            self.first_window = False    
        
        # Call solve function
        [self.R_t_x_sol,
        self.R_t_y_sol,
        self.R_t_z_sol,
        self.p_obj_sol,
        self.i1dot_sol,
        self.i1_sol,
        self.i2_sol,
        self.i1ddot_sol,
        self.i2dot_sol,
        self.i3_sol] = self.ocp_to_function(
        stepsize, 
        measured_positions.T,
        self.R_t_x_sol,
        self.R_t_y_sol,
        self.R_t_z_sol,
        self.p_obj_sol,
        self.i1dot_sol,
        self.i1_sol,
        self.i2_sol,
        self.i1ddot_sol[:,-1],
        self.i2dot_sol[:,-1],
        self.i3_sol[:,-1],)
                    
        invariants = np.array(np.vstack((self.i1_sol,self.i2_sol,self.i3_sol))).T
        calculated_trajectory = np.array(self.p_obj_sol).T
        calculated_movingframe = np.reshape(np.hstack((self.R_t_x_sol[:], self.R_t_y_sol[:], self.R_t_z_sol[:])), (-1,3,3)) # make a N x 3 x 3 array


        return invariants, calculated_trajectory, calculated_movingframe

        
    #%%     
    # def calculate_invariants_global_old(self,trajectory_meas,stepsize):
    #     """
    #     !! OLD AND OUTDATED !!
    #     """

    #     if trajectory_meas.shape[1] == 3:
    #         measured_positions = trajectory_meas
    #     else:
    #         measured_positions = trajectory_meas[:,:3,3]
    #     N = self.N 
    #     [ex,ey,ez] = ocp_helper.estimate_initial_frames(measured_positions,N)


    #     '''Initialization and setting values parameters'''
    #     start_time = time.time()
        
    #     print('initializing states')
    #     # Initialize states
    #     self.ocp.set_initial(self.R_t_x, ex.T)
    #     self.ocp.set_initial(self.R_t_y, ey.T)
    #     self.ocp.set_initial(self.R_t_z, ez.T)
    #     self.ocp.set_initial(self.p_obj, measured_positions[:,:].T)
    #     self.ocp.set_initial(self.i1dot,1e-12)
    #     self.ocp.set_initial(self.i1,1e0)
    #     self.ocp.set_initial(self.i2,1e-1)
        
    #     print('done with states')
        
    #     print('initializing controls')
    #     # Initialize controls
    #     self.ocp.set_initial(self.i1ddot,1e-12)
    #     self.ocp.set_initial(self.i2dot,1e-12)
    #     self.ocp.set_initial(self.i3,1e-12)
    #     print('done with controls')
        
    #     print('initializing parameters')
    #     # Set values parameters
    #     #self.ocp.set_value(self.R_t_0, np.eye(3,3))
    #     self.ocp.set_value(self.p_obj_m, measured_positions[0:-1,:].T)       
    #     self.ocp.set_value(self.h,stepsize)
    #     print('done with parameters')
        
    #     end_time = time.time()
    #     print('')
    #     print("Initialization: ")
    #     print(end_time - start_time)
        
    #     # Solve the NLP
    #     start_time = time.time()
    #     sol = self.ocp.solve()
    #     end_time = time.time()
    #     print('')
    #     print("Solving: ")
    #     print(end_time - start_time)
        
    #     self.sol = sol
        
    #     start_time = time.time()

    #     t1,i1 = sol.sample(self.i1,grid='control')
    #     _,i2 = sol.sample(self.i2,grid='control')
    #     _,i3 = sol.sample(self.i3,grid='control')
    #     invariants = np.array((i1,i2,i3)).T
    #    # invariants = np.vstack([invariants,invariants[-1,:]])
        
    #     _,calculated_trajectory = sol.sample(self.p_obj,grid='control')
    #     _,calculated_movingframe = sol.sample(self.R_t_x,grid='control')
    #     # R_t = np.hstack ( R_tx, Rty, Rtz)
    #     end_time = time.time()
    #     print('')
    #     print("sampling solution: ")
    #     print(end_time - start_time)
        
    #     return invariants, calculated_trajectory, calculated_movingframe

    # def calculate_invariants_online_old(self,measured_positions,stepsize,sample_jump=1):
    #     """
    #     !! OLD AND OUTDATED !!
    #     """
        
    #     if self.first_window:
    #         # Calculate invariants in first window
    #         invariants, calculated_trajectory, calculated_movingframe = self.calculate_invariants_global_old(measured_positions,stepsize)
    #         self.first_window = False
                    
    #         #self.opti.subject_to( self.R_t[0] == self.R_t_0 ) # Add continuity constraints on first sample
    #         #self.opti.subject_to( self.p_obj[0] == self.p_obj_m[0]) # Add continuity constraints on first sample
            
    #         return invariants, calculated_trajectory, calculated_movingframe
    #     else:
       
    #         # Set values parameters
    #         self.ocp.set_value(self.p_obj_m, measured_positions[0:-1,:].T)
    #         self.ocp.set_value(self.h,stepsize);
    #         #self.opti.set_value(self.R_t_0, self.sol.value(self.R_t[sample_jump])); # continuity moving frame
    #         #self.opti.set_value(self.p_obj_m[0], self.sol.value(self.p_obj[sample_jump])); # continuity position
            
    #         # Solve and post-process
    #         sol = self.ocp.solve_limited()
    #         self.sol = sol
            
    #         # Extract the solved variables
    #         _,i1 = sol.sample(self.i1,grid='control')
    #         _,i2 = sol.sample(self.i2,grid='control')
    #         _,i3 = sol.sample(self.i3,grid='control')
    #         invariants = np.array((i1,i2,i3)).T            
    #         _,calculated_trajectory = sol.sample(self.p_obj,grid='control')
    #         _,calculated_movingframe = sol.sample(self.R_t_x,grid='control')

    #     # Initialize states and controls
    #     self.ocp.set_initial(self.R_t_x, self.sol.sample(self.R_t_x,grid='control')[1].T)
    #     self.ocp.set_initial(self.R_t_y, self.sol.sample(self.R_t_y,grid='control')[1].T)
    #     self.ocp.set_initial(self.R_t_z, self.sol.sample(self.R_t_z,grid='control')[1].T)
    #     self.ocp.set_initial(self.p_obj, self.sol.sample(self.p_obj,grid='control')[1].T)
    #     self.ocp.set_initial(self.i1dot, self.sol.sample(self.i1dot,grid='control')[1].T)
    #     self.ocp.set_initial(self.i1,    self.sol.sample(self.i1,grid='control')[1].T)
    #     self.ocp.set_initial(self.i2,    self.sol.sample(self.i2,grid='control')[1].T)
    #     self.ocp.set_initial(self.i1ddot,self.sol.sample(self.i1ddot,grid='control')[1].T)
    #     self.ocp.set_initial(self.i2dot, self.sol.sample(self.i2dot,grid='control')[1].T)
    #     self.ocp.set_initial(self.i3,    self.sol.sample(self.i3,grid='control')[1].T)

    #     return invariants, calculated_trajectory, calculated_movingframe

if __name__ == "__main__":
    # Example data for measured positions and the stepsize
    measured_positions = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12], [13, 14, 15], [16, 17, 18]])
    stepsize = 0.05

    # Test the functionalities of the class
    OCP = OCP_calc_pos(nb_samples=np.size(measured_positions,0),fatrop_solver=False)
 
    # Call the calculate_invariants_online function
    calc_invariants, calc_trajectory, calc_movingframes = OCP.calculate_invariants_online(measured_positions, stepsize)

    # Print the results
    print("Calculated invariants:")
    print(calc_invariants)
    print("Calculated Moving Frame:")
    print(calc_movingframes)
    print("Calculated Trajectory:")
    print(calc_trajectory)


"""
Contains the class implementataion of the GJK Algorithm

Script taken from: https://github.com/mohith-sakthivel/Fast_GJK-collision_detection_bottle
"""

import numpy as np
# np.random.seed(0)


class ComputeDist():
    def __init__(self, ObjectList=[], tolerance=0.00001):
        """
        Initialize the cobject environment
        """
        self.ObjCount = 0
        self.ObjID = {}
        self.ConvObject = []
        self.tolerance = tolerance
        for shape in ObjectList:
            assert len(shape) == 2, \
                "Shape name and a shape class object should be provided."
            assert type(shape[0]) == str, "Shape name must be a string"
            self.ObjID[shape[0]] = self.ObjCount
            self.ConvObject.append(shape[1])
            self.ObjCount += 1
        self.WitnessPts = {}

    def SupportMinkowskiDiff(self, Shape1, Shape2, Dir):
        """
        Calculate the support of the Minkowski difference Shape1-Shape2
        """
        Pt1, Dist1, Index1 = Shape1.SupportFunc(Dir)
        Pt2, Dist2, Index2 = Shape2.SupportFunc(-Dir)
        return np.array([Pt1-Pt2, Dist1+Dist2, (Index1, Index2)], dtype=object)

    def S1D(self, Simplex):
        """
        Sub-routine for 1-simplex.
        Searches the voronoi regions of a line.
        """
        if all(Simplex[1] == Simplex[0]):
            return Simplex[1:], np.ones(1, dtype='float64')
        t = Simplex[1]-Simplex[0]
        po = -(np.dot(Simplex[1], t)/np.dot(t, t))*t+Simplex[1]
        u_max = 0
        # Find on which axis to project the simplex to avoid degeneracy
        for i in range(3):
            u = Simplex[0][i]-Simplex[1][i]
            if abs(u) > abs(u_max):
                u_max = u
                Index = i
        # Only i-th co-ordinate is retained
        k = 1
        C2 = np.zeros(2, dtype='float64')
        # Check if the origin is on the line segment
        for j in range(2):
            C2[j] = ((-1)**(j+1))*(Simplex[k][Index]-po[Index])
            k = j
        if (u_max > 0 and all(C2 > 0)) or (u_max < 0 and all(C2 < 0)):
            return Simplex, C2/u_max
        else:
            # Find which end point of the line segment is closest to the origin
            if (np.linalg.norm(Simplex[0]) < np.linalg.norm(Simplex[1])):
                return Simplex[0:1], np.ones(1, dtype='float64')
            else:
                return Simplex[1:], np.ones(1, dtype='float64')

    def S2D(self, Simplex):
        """
        Sub-routine for 2-simplex.
        Searches the voronoi regions of a planar triangle.
        """
        n = np.cross(Simplex[1]-Simplex[0], Simplex[2]-Simplex[0])
        po = (np.dot(Simplex[0], n)/np.dot(n, n))*n

        # Find on which plane to project the simplex to avoid degeneracy
        u_max = 0
        k = 1
        l = 2
        for i in range(0, 3):
            u = ((-1)**(i))*(
                             Simplex[0][k]*Simplex[1][l]
                             + Simplex[1][k]*Simplex[2][l]
                             + Simplex[2][k]*Simplex[0][l]
                             - Simplex[1][k]*Simplex[0][l]
                             - Simplex[2][k]*Simplex[1][l]
                             - Simplex[0][k]*Simplex[2][l])
            if abs(u) > abs(u_max):
                u_max = u
                J = i
            k = l
            l = i
        # co-ordinate J is discarded
        x, y = np.delete(np.arange(0, 3), J)
        k = 1
        l = 2
        C3 = np.zeros(3, dtype='float64')
        # Check if the origin is within the triangle
        for j in range(3):
            C3[j] = (
                     po[x]*Simplex[k][y]+po[y]*Simplex[l][x]
                     + Simplex[k][x]*Simplex[l][y]
                     - po[x]*Simplex[l][y]-po[y]*Simplex[k][x]
                     - Simplex[l][x]*Simplex[k][y])
            k = l
            l = j
        if (u_max > 0 and all(C3 > 0)) or (u_max < 0 and all(C3 < 0)):
            return Simplex, C3/u_max
        d = np.Infinity
        # Find which side of the triangle is closest to the origin
        for j in range(0, 3):
            if (u_max >= 0 and -C3[j] >= 0) or (u_max <= 0 and -C3[j] <= 0):
                Simplex1D = np.delete(Simplex, j, axis=0)
                W_astrix, Lambda_astrix = self.S1D(Simplex1D)
                d_astrix = np.linalg.norm(np.matmul(Lambda_astrix, W_astrix))
                if d_astrix < d:
                    W = W_astrix
                    Lamda = Lambda_astrix
                    d = d_astrix
        return W, Lamda

    def S3D(self, Simplex):
        """
        Sub-routine for 3-simplex.
        Searches the voronoi regions of a tetrahedron.
        """
        M = np.vstack([Simplex.T, np.ones((1, 4), dtype='float64')])
        detM = 0
        C4 = np.zeros(4, dtype='float64')
        # Check if the origin is within the tetrahedron
        for j in range(4):
            C4[j] = ((-1)**((j+1)+4))*np.linalg.det(
                            np.hstack([M[0:3, 0:j], M[0:3, j+1:]]))
            detM += C4[j]
        if (detM > 0 and all(C4 > 0)) or (detM < 0 and all(C4 < 0)):
            return Simplex, C4/detM
        d = np.Infinity
        # Find which face of the tetrahedron is closest to the origin
        for j in range(0, 4):
            if (detM >= 0 and -C4[j] >= 0) or (detM <= 0 and -C4[j] <= 0):
                Simplex2D = np.delete(Simplex, j, axis=0)
                W_astrix, Lambda_astrix = self.S2D(Simplex2D)
                d_astrix = np.linalg.norm(np.matmul(Lambda_astrix, W_astrix))
                if d_astrix < d:
                    W = W_astrix
                    Lamda = Lambda_astrix
                    d = d_astrix
        return W, Lamda

    def SignedVolumes(self, Simplex):
        """
        Performs the signed volumes distance algorithm
        on the simplex specified
        """
        # Call routine based on simplex size
        if len(Simplex) == 4:
            return self.S3D(Simplex)
        elif len(Simplex) == 3:
            return self.S2D(Simplex)
        elif len(Simplex) == 2:
            return self.S1D(Simplex)
        elif len(Simplex) == 1:
            return Simplex, np.ones((1), dtype='float64')
        else:
            print("error")

    def GetDist(self, ShapeID1, ShapeID2):
        """
        Computes the distance between 2 convex bodies
        """
        k = 0
        Shape1 = self.ObjID[ShapeID1]
        Shape2 = self.ObjID[ShapeID2]
        # Initialize the simplex
        # Select a random initial point for the simplex
        index_1 = np.random.randint(0, self.ConvObject[Shape1].NoOfVertices)
        index_2 = np.random.randint(0, self.ConvObject[Shape2].NoOfVertices)
        InitialPt = (
                     self.ConvObject[Shape1].CurrentVertices[index_1] -
                     self.ConvObject[Shape2].CurrentVertices[index_2])

        Simplex = InitialPt.reshape((1, 3))
        # Set initial direction opposite to the selected point to maximize area
        NewPt, _, _ = self.SupportMinkowskiDiff(
                                                self.ConvObject[Shape1],
                                                self.ConvObject[Shape2],
                                                -InitialPt)

        Simplex = np.vstack([NewPt, Simplex])
        while True:
            k += 1
            # Use signed volumes to get the supporting points and their weights
            Simplex, Lambda = self.SignedVolumes(Simplex)
            vk = np.matmul(Lambda, Simplex)
            NewPt, hk, _ = self.SupportMinkowskiDiff(
                                                     self.ConvObject[Shape1],
                                                     self.ConvObject[Shape2],
                                                     -vk)
            vk_square = np.dot(vk, vk)
            gk = vk_square+hk
            if (gk < self.tolerance or len(Simplex) == 4):
                return np.sqrt(vk_square)
            Simplex = np.vstack([NewPt, Simplex])





import numpy as np
import matplotlib.pyplot as plt
import invariants_py.reparameterization as reparam
import scipy.interpolate as ip
from invariants_py.opti_calculate_vector_invariants_position_mj import OCP_calc_pos
from IPython.display import clear_output
import invariants_py.data_handler as dh

#%%
data_location = dh.find_data_path('contour_coordinates.out')
position_data = np.loadtxt(data_location, dtype='float')
trajectory,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_positiontrajectory_arclength(position_data)
stepsize_orig = stepsize
arclength_n = arclength/arclength[-1]

plt.figure(figsize=(8,3))
plt.axis('equal')
plt.plot(trajectory[:,0],trajectory[:,1],'.-')

#%%
# specify optimization problem symbolically
FS_calculation_problem = OCP_calc_pos(window_len=nb_samples, w_pos = 100, w_regul = 10**-9)

# calculate invariants given measurements
invariants, calculate_trajectory, movingframes = FS_calculation_problem.calculate_invariants_global(trajectory,stepsize)

#%%
plt.figure(figsize=(8,3))
plt.axis('equal')
plt.plot(trajectory[:,0],trajectory[:,1],'.-')
plt.plot(calculate_trajectory[:,0],calculate_trajectory[:,1],'.-')

f, (ax1, ax2, ax3) = plt.subplots(1, 3, sharey=True, figsize=(8,3))
ax1.plot(arclength,invariants[:,0])
ax1.set_title('Velocity [m/-]')
ax2.plot(arclength,invariants[:,1])
ax2.set_title('Curvature [rad/-]')
ax3.plot(arclength,invariants[:,2])
ax3.set_title('Torsion [rad/-]')

plt.figure()
plt.plot(invariants[:,0])
plt.figure()
plt.plot(invariants[:,1])
plt.figure()
plt.plot(invariants[:,2])


#%%
# Spline of model
knots = np.concatenate(([arclength[0]],[arclength[0]],arclength,[arclength[-1]],[arclength[-1]]))
degree = 3
spline_model_trajectory = ip.BSpline(knots,trajectory,degree)

def simulate_noisy_measurements(model_trajectory, current_progress, stepsize, online_window_len):
    
    noise_std = 0.000
    
    progress_values = np.linspace(current_progress, current_progress-online_window_len*stepsize, online_window_len )
    #print(progress_values)
    noisy_measurements = np.array([model_trajectory(i) for i in progress_values]) 

    return noisy_measurements + np.random.randn(online_window_len,3)*noise_std


#%%
test_measurements = simulate_noisy_measurements(spline_model_trajectory,0.8,0.005,20)

plt.figure(figsize=(8,3))
plt.axis('equal')
plt.plot(trajectory[:,0],trajectory[:,1],'.-')
plt.plot(test_measurements[:,0],test_measurements[:,1],'k.-')


#%%
window_len = 20
stepsize = 0.005
window_increment = 10

# specify optimization problem symbolically
FS_online_calculation_problem = OCP_calc_pos(window_len=window_len, w_pos = 100, w_regul = 10**-9)

# Visualization
current_progress = 0.0 + window_len*stepsize

# Create figure
plt.figure(figsize=(14,6))

while current_progress <= arclength[-1]:

    #print(f"current progress = {current_progress}")
    
    measurements = simulate_noisy_measurements(spline_model_trajectory,current_progress,stepsize,window_len)

    # Calculate invariants in window
    invariants_online, trajectory_online, mf = FS_online_calculation_problem.calculate_invariants_online(measurements,stepsize,window_increment)

    # Visualization
    xvector = np.linspace(current_progress-window_len*stepsize, current_progress , window_len)
    
    clear_output(wait=True)
    
    plt.subplot(2,2,1)
    plt.plot(trajectory[:,0],trajectory[:,1],'.-')
    plt.plot(measurements[:,0],measurements[:,1],'k.')
    plt.plot(trajectory_online[:,0],trajectory_online[:,1])
    
    plt.subplot(2,2,3)
    plt.plot(arclength,invariants[:,0])
    plt.plot(xvector,abs(invariants_online[:,0]))
    plt.plot(0,0)
    plt.title('Velocity [m/-]')
    
    plt.subplot(2,2,2)
    plt.plot(arclength,abs(invariants[:,1]))
    plt.plot(xvector,abs(invariants_online[:,1]))
    plt.plot(0,0)
    plt.title('Curvature [rad/-]')
    
    plt.subplot(2,2,4)
    plt.plot(arclength,invariants[:,2])
    plt.plot(xvector,abs(invariants_online[:,2]))
    plt.plot(0,1)
    plt.title('Torsion [rad/-]')

    if plt.get_backend() != 'agg':
        error()
        plt.show()
    
    
    current_progress = round(current_progress + window_increment*stepsize,3) # start index next window





# Imports
import numpy as np
import matplotlib.pyplot as plt
from invariants_py.rockit_generate_rotation_from_vector_invariants import OCP_gen_rot as FS_gen_rot
from scipy.spatial.transform import Rotation as R
from invariants_py.orthonormalize_rotation import orthonormalize_rotation as orthonormalize
import invariants_py.plotters as pl
from invariants_py.reparameterization import interpR
import pickle
import invariants_py.SO3 as SO3
import os
from invariants_py import data_handler as dh
plt.close('all')
# define class for OCP results
class OCP_results:

    def __init__(self,FSt_frames,FSr_frames,Obj_pos,Obj_frames,invariants):
        self.FSt_frames = FSt_frames
        self.FSr_frames = FSr_frames
        self.Obj_pos = Obj_pos
        self.Obj_frames = Obj_frames
        self.invariants = invariants
        
# Read dictionary pkl file
data_location = dh.find_data_path('inv_model.pkl')
with open(data_location, 'rb') as fp:
    optim_calc_results = pickle.load(fp)

number_samples = 100
progress_values = np.linspace(0, 1, number_samples)
model_invariants = optim_calc_results.invariants
#model_invariants[:,1] = -optim_calc_results.invariants[:,1]
new_stepsize = progress_values[1] - progress_values[0] 

# new constraints
current_index = 0
R_obj_start = orthonormalize(optim_calc_results.Obj_frames[current_index])
# FSr_start = orthonormalize(optim_calc_results.FSr_frames[current_index])
alpha = 150
rotate = R.from_euler('z', alpha, degrees=True)
R_obj_end =  orthonormalize(rotate.as_matrix() @ optim_calc_results.Obj_frames[-1])
# FSr_end = orthonormalize(optim_calc_results.FSr_frames[-1])

# define new class for OCP results
optim_gen_results = OCP_results(FSt_frames = [], FSr_frames = [], Obj_pos = [], Obj_frames = [], invariants = np.zeros((number_samples,6)))

# Linear initialization
R_obj_init = interpR(np.linspace(0, 1, len(optim_calc_results.Obj_frames)), [0,1], np.array([R_obj_start, R_obj_end]))
# R_r_init = interpR(np.linspace(0, 1, len(optim_calc_results.FSr_frames)), [0,1], np.array([FSr_start, FSr_end]))

skew_angle = SO3.logm(R_obj_start.T @ R_obj_end)
angle_vec_in_body = np.array([skew_angle[2,1],skew_angle[0,2],skew_angle[1,0]])
angle_vec_in_world = R_obj_start@angle_vec_in_body
angle_norm = np.linalg.norm(angle_vec_in_world)
invars_init = np.tile(np.array([angle_norm,0.001,0.001]),(number_samples-1,1))
e_x_fs_init = angle_vec_in_world/angle_norm
e_y_fs_init = [0,1,0]
e_y_fs_init = e_y_fs_init - np.dot(e_y_fs_init,e_x_fs_init)*e_x_fs_init
e_y_fs_init = e_y_fs_init/np.linalg.norm(e_y_fs_init)
e_z_fs_init = np.cross(e_x_fs_init,e_y_fs_init)
R_fs_init = np.array([e_x_fs_init,e_y_fs_init,e_z_fs_init]).T

R_fs_init_array = []
for k in range(number_samples):
    R_fs_init_array.append(R_fs_init)
R_fs_init_array = np.array(R_fs_init_array)

boundary_constraints = {"orientation": {"initial": R_obj_start, "final": R_obj_end}, "moving-frame-orientation": {"initial": R_fs_init, "final": R_fs_init}}

initial_values = {"invariants-orientation": invars_init, "trajectory-orientation": R_obj_init, "moving-frame-orientation": R_fs_init_array}

# specify optimization problem symbolically
FS_online_generation_problem_rot = FS_gen_rot(boundary_constraints, number_samples, fatrop_solver = 1)

# Solve
optim_gen_results.invariants[:,:3], optim_gen_results.Obj_frames, optim_gen_results.FSr_frames, tot_time_rot = FS_online_generation_problem_rot.generate_trajectory(model_invariants[:,:3],boundary_constraints,new_stepsize,initial_values=initial_values)


for i in range(len(optim_gen_results.Obj_frames)):
    optim_gen_results.Obj_frames[i] = orthonormalize(optim_gen_results.Obj_frames[i])

fig = plt.figure()
plt.plot(optim_gen_results.invariants[:,0])

fig99 = plt.figure(figsize=(14,8))
ax99 = fig99.add_subplot(111, projection='3d')
opener_location = dh.find_data_path('opener.stl')
pl.plot_stl(opener_location,[0,0,0],optim_calc_results.Obj_frames[-1],colour="r",alpha=0.5,ax=ax99)
pl.plot_stl(opener_location,[0,0,0],R_obj_end,colour="b",alpha=0.5,ax=ax99)
pl.plot_stl(opener_location,[0,0,0],optim_gen_results.Obj_frames[-1],colour="g",alpha=0.5,ax=ax99)


print(np.transpose(R_obj_end) @ optim_gen_results.Obj_frames[-1] - np.eye(3,3))
assert(np.linalg.norm(np.transpose(R_obj_end) @ optim_gen_results.Obj_frames[-1] - np.eye(3,3)) < 1.0)